"""
Wrappers around Google APIs.
"""
import json
import os
from base64 import b64encode
from collections import OrderedDict
from collections import namedtuple
from datetime import datetime
from datetime import timedelta
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

try:
    # this is only an issue with Python 2.7 and if the
    # Google-API packages were not installed with msl-io
    from enum import Enum
except ImportError:
    Enum = object

# having the Google-API packages are optional
try:
    from google.auth.exceptions import RefreshError
    from google.auth.transport.requests import Request
    from google.oauth2.credentials import Credentials
    from google_auth_oauthlib.flow import InstalledAppFlow
    from googleapiclient.discovery import build
    from googleapiclient.errors import HttpError
    from googleapiclient.http import DEFAULT_CHUNK_SIZE
    from googleapiclient.http import MediaFileUpload
    from googleapiclient.http import MediaIoBaseDownload
    HAS_GOOGLE_API = True
except ImportError:
    DEFAULT_CHUNK_SIZE = 100 * 1024 * 1024
    HAS_GOOGLE_API = False

from .constants import HOME_DIR
from .constants import IS_PYTHON2


def _authenticate(token, client_secrets_file, scopes):
    """Authenticate with a Google API.

    Parameters
    ----------
    token : :class:`str`
        The path to a token file. If it does not exist then it will be created.
    client_secrets_file : :class:`str`
        The "client secrets" file to use to generate the OAuth credentials.
    scopes : :class:`list` of :class:`str`
        The list of scopes to enable.

    Returns
    -------
    :class:`google.oauth2.credentials.Credentials`
        The OAuth 2.0 credentials for the user.
    """
    if not HAS_GOOGLE_API:
        raise RuntimeError(
            'You must install the Google-API packages, run\n'
            '  pip install google-api-python-client google-auth-httplib2 google-auth-oauthlib'
        )

    credentials = None

    # load the token from an environment variable if it exists
    # ignore the '.json' extension
    token_env_name = os.path.basename(token)[:-5].replace('-', '_').upper()
    if token_env_name in os.environ:
        info = json.loads(os.environ[token_env_name])
        credentials = Credentials.from_authorized_user_info(info, scopes=scopes)

    # load the cached token file if it exists
    if not credentials and os.path.isfile(token):
        credentials = Credentials.from_authorized_user_file(token, scopes=scopes)

    # if there are no (valid) credentials available then let the user log in
    if not credentials or not credentials.valid:
        if credentials and credentials.expired and credentials.refresh_token:
            try:
                credentials.refresh(Request())
            except RefreshError as err:
                if os.path.isfile(token) and not os.getenv('MSL_IO_RUNNING_TESTS'):
                    message = '{}: {}\nDo you want to delete the token file and re-authenticate ' \
                              '(y/N)? '.format(err.__class__.__name__, err.args[0])
                    if IS_PYTHON2:
                        yes_no = raw_input(message)
                    else:
                        yes_no = input(message)
                    if yes_no.lower().startswith('y'):
                        os.remove(token)
                        return _authenticate(token, client_secrets_file, scopes)
                raise
        else:
            if not client_secrets_file:
                raise OSError('You must specify the path to a "client secrets" file as the credentials')
            flow = InstalledAppFlow.from_client_secrets_file(client_secrets_file, scopes)
            credentials = flow.run_local_server(port=0)

        # save the credentials for the next run
        if token_env_name in os.environ:
            os.environ[token_env_name] = credentials.to_json()
        else:
            # make sure that all parent directories exist before creating the file
            dirname = os.path.dirname(token)
            if dirname and not os.path.isdir(dirname):
                os.makedirs(dirname)
            with open(token, mode='wt') as fp:
                fp.write(credentials.to_json())

    return credentials


class GoogleAPI(object):

    def __init__(self, service, version, credentials, scopes, read_only, account):
        """Base class for all Google APIs."""

        name = '{}-'.format(account) if account else ''
        readonly = '-readonly' if read_only else ''
        filename = '{}token-{}{}.json'.format(name, service, readonly)
        token = os.path.join(HOME_DIR, filename)
        oauth = _authenticate(token, credentials, scopes)
        self._service = build(service, version, credentials=oauth)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def service(self):
        """The Resource object with methods for interacting with the API service."""
        return self._service

    def close(self):
        """Close the connection to the API service."""
        self._service.close()


class GDrive(GoogleAPI):

    MIME_TYPE_FOLDER = 'application/vnd.google-apps.folder'
    ROOT_NAMES = ['Google Drive', 'My Drive', 'Drive']

    def __init__(self, account=None, credentials=None, read_only=True, scopes=None):
        """Interact with Google Drive.

        .. attention::
           You must follow the instructions in the prerequisites section for setting up the
           `Drive API <https://developers.google.com/drive/api/quickstart/python#prerequisites>`_
           before you can use this class. It is also useful to be aware of the
           `refresh token expiration <https://developers.google.com/identity/protocols/oauth2#expiration>`_
           policy.

        .. _Media type: https://www.iana.org/assignments/media-types/media-types.xhtml
        .. _Drive MIME type: https://developers.google.com/drive/api/guides/mime-types

        Parameters
        ----------
        account : :class:`str`, optional
            Since a person may have multiple Google accounts, and multiple people
            may run the same code, this parameter decides which token to load
            to authenticate with the Google API. The value can be any text (or
            :data:`None`) that you want to associate with a particular Google
            account, provided that it contains valid characters for a filename.
            The value that you chose when you authenticated with your `credentials`
            should be used for all future instances of this class to access that
            particular Google account. You can associate a different value with
            a Google account at any time (by passing in a different `account`
            value), but you will be asked to authenticate with your `credentials`
            again, or, alternatively, you can rename the token files located in
            :const:`~msl.io.constants.HOME_DIR` to match the new `account` value.
        credentials : :class:`str`, optional
            The path to the `client secrets` OAuth credential file. This
            parameter only needs to be specified the first time that you
            authenticate with a particular Google account or if you delete
            the token file that was created when you previously authenticated.
        read_only : :class:`bool`, optional
            Whether to interact with Google Drive in read-only mode.
        scopes : :class:`list` of :class:`str`, optional
            The list of scopes to enable for the Google API. See
            `Drive scopes <https://developers.google.com/identity/protocols/oauth2/scopes#drive>`_
            for more details. If not specified then default scopes are chosen
            based on the value of `read_only`.
        """
        if not scopes:
            if read_only:
                scopes = [
                    'https://www.googleapis.com/auth/drive.readonly',
                    'https://www.googleapis.com/auth/drive.metadata.readonly'
                ]
            else:
                scopes = [
                    'https://www.googleapis.com/auth/drive',
                    'https://www.googleapis.com/auth/drive.metadata',
                ]

        super(GDrive, self).__init__(
            'drive', 'v3', credentials, scopes, read_only, account)

        self._files = self._service.files()
        self._drives = self._service.drives()

    @staticmethod
    def _folder_hierarchy(folder):
        # create a list of sub-folder names in the folder hierarchy
        f = folder
        names = []
        while True:
            f, name = os.path.split(f)
            if not name or name in GDrive.ROOT_NAMES:
                break
            names.append(name)
        return names[::-1]

    def folder_id(self, folder, parent_id=None):
        """Get the ID of a Google Drive folder.

        Parameters
        ----------
        folder : :class:`str`
            The path to a Google Drive file.
        parent_id : :class:`str`, optional
            The ID of the parent folder that `folder` is relative to. If not
            specified then `folder` is relative to the `My Drive` root folder.
            If `folder` is in a `Shared drive` then you must specify the
            ID of a parent folder.

        Returns
        -------
        :class:`str`
            The folder ID.
        """
        folder_id = parent_id or 'root'
        names = GDrive._folder_hierarchy(folder)
        for name in names:
            q = '"{}" in parents and name="{}" and trashed=false and mimeType="{}"'.format(
                folder_id, name, GDrive.MIME_TYPE_FOLDER
            )
            response = self._files.list(
                q=q,
                fields='files(id,name)',
                includeItemsFromAllDrives=True,
                supportsAllDrives=True,
            ).execute()
            files = response['files']
            if not files:
                raise OSError('Not a valid Google Drive folder {!r}'.format(folder))
            if len(files) > 1:
                matches = '\n  '.join(str(file) for file in files)
                raise OSError('Multiple folders exist for {!r}\n  {}'.format(name, matches))

            first = files[0]
            assert name == first['name'], '{!r} != {!r}'.format(name, first['name'])
            folder_id = first['id']

        return folder_id

    def file_id(self, file, mime_type=None, folder_id=None):
        """Get the ID of a Google Drive file.

        Parameters
        ----------
        file : :class:`str`
            The path to a Google Drive file.
        mime_type : :class:`str`, optional
            The `Drive MIME type`_ or `Media type`_ to use to filter the results.
        folder_id : :class:`str`, optional
            The ID of the folder that `file` is relative to. If not specified
            then `file` is relative to the `My Drive` root folder.
            If `file` is in a `Shared drive` then you must specify the
            ID of a parent folder.

        Returns
        -------
        :class:`str`
            The file ID.
        """
        folders, name = os.path.split(file)
        folder_id = self.folder_id(folders, parent_id=folder_id)

        q = '"{}" in parents and name="{}" and trashed=false'.format(folder_id, name)
        if not mime_type:
            q += ' and mimeType!="{}"'.format(GDrive.MIME_TYPE_FOLDER)
        else:
            q += ' and mimeType="{}"'.format(mime_type)

        response = self._files.list(
            q=q,
            fields='files(id,name,mimeType)',
            includeItemsFromAllDrives=True,
            supportsAllDrives=True,
        ).execute()
        files = response['files']
        if not files:
            raise OSError('Not a valid Google Drive file {!r}'.format(file))
        if len(files) > 1:
            mime_types = '\n  '.join(f['mimeType'] for f in files)
            raise OSError('Multiple files exist for {!r}. '
                          'Filter by MIME type:\n  {}'.format(file, mime_types))

        first = files[0]
        assert name == first['name'], '{!r} != {!r}'.format(name, first['name'])
        return first['id']

    def is_file(self, file, mime_type=None, folder_id=None):
        """Check if a file exists.

        Parameters
        ----------
        file : :class:`str`
            The path to a Google Drive file.
        mime_type : :class:`str`, optional
            The `Drive MIME type`_ or `Media type`_ to use to filter the results.
        folder_id : :class:`str`, optional
            The ID of the folder that `file` is relative to. If not specified
            then `file` is relative to the `My Drive` root folder.
            If `file` is in a `Shared drive` then you must specify the
            ID of a parent folder.

        Returns
        -------
        :class:`bool`
            Whether the file exists.
        """
        try:
            self.file_id(file, mime_type=mime_type, folder_id=folder_id)
        except OSError as err:
            return str(err).startswith('Multiple files')
        else:
            return True

    def is_folder(self, folder, parent_id=None):
        """Check if a folder exists.

        Parameters
        ----------
        folder : :class:`str`
            The path to a Google Drive folder.
        parent_id : :class:`str`, optional
            The ID of the parent folder that `folder` is relative to. If not
            specified then `folder` is relative to the `My Drive` root folder.
            If `folder` is in a `Shared drive` then you must specify the
            ID of a parent folder.

        Returns
        -------
        :class:`bool`
            Whether the folder exists.
        """
        try:
            self.folder_id(folder, parent_id=parent_id)
        except OSError as err:
            return str(err).startswith('Multiple folders')
        else:
            return True

    def create_folder(self, folder, parent_id=None):
        """Create a folder.

        Makes all intermediate-level folders needed to contain the leaf directory.

        Parameters
        ----------
        folder : :class:`str`
            The folder(s) to create, for example, ``'folder1'`` or
            ``'folder1/folder2/folder3'``.
        parent_id : :class:`str`, optional
            The ID of the parent folder that `folder` is relative to. If not
            specified then `folder` is relative to the `My Drive` root folder.
            If `folder` is in a `Shared drive` then you must specify the
            ID of a parent folder.

        Returns
        -------
        :class:`str`
            The ID of the last (right most) folder that was created.
        """
        names = GDrive._folder_hierarchy(folder)
        response = {'id': parent_id or 'root'}
        for name in names:
            request = self._files.create(
                body={
                    'name': name,
                    'mimeType': GDrive.MIME_TYPE_FOLDER,
                    'parents': [response['id']],
                },
                fields='id',
                supportsAllDrives=True,
            )
            response = request.execute()
        return response['id']

    def delete(self, file_or_folder_id):
        """Delete a file or a folder.

        Files that are in read-only mode cannot be deleted.

        .. danger::
           Permanently deletes the file or folder owned by the user without
           moving it to the trash. If the target is a folder, then all files
           and sub-folders contained within the folder (that are owned by the
           user) are also permanently deleted.

        Parameters
        ----------
        file_or_folder_id : :class:`str`
            The ID of the file or folder to delete.
        """
        if self.is_read_only(file_or_folder_id):
            # The API allows for a file to be deleted if it is in read-only mode,
            # but we will not allow it to be deleted
            raise RuntimeError('Cannot delete the file since it is in read-only mode')

        self._files.delete(
            fileId=file_or_folder_id,
            supportsAllDrives=True,
        ).execute()

    def empty_trash(self):
        """Permanently delete all files in the trash."""
        self._files.emptyTrash().execute()

    def upload(self, file, folder_id=None, mime_type=None, resumable=False, chunk_size=DEFAULT_CHUNK_SIZE):
        """Upload a file.

        Parameters
        ----------
        file : :class:`str`
            The file to upload.
        folder_id : :class:`str`, optional
            The ID of the folder to upload the file to. If not specified then
            uploads to the `My Drive` root folder.
        mime_type : :class:`str`, optional
            The `Drive MIME type`_ or `Media type`_ of the file
            (e.g., ``'text/csv'``). If not specified then a type will be
            guessed based on the file extension.
        resumable : :class:`bool`
            Whether the upload can be resumed.
        chunk_size : :class:`int`
            The file will be uploaded in chunks of this many bytes. Only used
            if `resumable` is :data:`True`. Pass in a value of -1 if the file
            is to be uploaded in a single chunk. Note that Google App Engine
            has a 5MB limit on request size, so you should never set
            `chunk_size` to be >5MB or to -1 (if the file size is >5MB).

        Returns
        -------
        :class:`str`
            The ID of the file that was uploaded.
        """
        parent_id = folder_id or 'root'
        filename = os.path.basename(file)

        body = {'name': filename, 'parents': [parent_id]}
        if mime_type:
            body['mimeType'] = mime_type

        request = self._files.create(
            body=body,
            media_body=MediaFileUpload(
                file,
                mimetype=mime_type,
                chunksize=chunk_size,
                resumable=resumable
            ),
            fields='id',
            supportsAllDrives=True,
        )
        response = request.execute()
        return response['id']

    def download(self, file_id, save_to=None, num_retries=0, chunk_size=DEFAULT_CHUNK_SIZE, callback=None):
        """Download a file.

        Parameters
        ----------
        file_id : :class:`str`
            The ID of the file to download.
        save_to : :term:`path-like <path-like object>` or :term:`file-like <file object>`, optional
            The location to save the file to. If a directory is specified
            then the file will be saved to that directory using the filename
            of the remote file. To save the file with a new filename, specify
            the new filename in `save_to`. Default is to save the file to the
            current working directory using the remote filename.
        num_retries : :class:`int`, optional
            The number of times to retry the download.
            If zero (default) then attempt the request only once.
        chunk_size : :class:`int`, optional
            The file will be downloaded in chunks of this many bytes.
        callback
            The callback to call after each chunk of the file is downloaded.
            The `callback` accepts one positional argument, for example::

                def handler(file):
                    print(file.progress(), file.total_size, file.resumable_progress)

                drive.download('0Bwab3C2ejYSdM190b2psXy1C50P', callback=handler)

        """
        if hasattr(save_to, 'write'):
            fh = save_to
        else:
            if not save_to or os.path.isdir(save_to):
                response = self._files.get(
                    fileId=file_id,
                    fields='name',
                    supportsAllDrives=True,
                ).execute()
                name = response['name']
                if save_to and os.path.isdir(save_to):
                    save_to = os.path.join(save_to, name)
                else:
                    save_to = name
            fh = open(save_to, mode='wb')

        request = self._files.get_media(fileId=file_id, supportsAllDrives=True)
        downloader = MediaIoBaseDownload(fh, request, chunksize=chunk_size)
        done = False
        while not done:
            status, done = downloader.next_chunk(num_retries=num_retries)
            if callback:
                callback(status)

        if fh is not save_to:  # then close the file that was opened
            fh.close()

    def path(self, file_or_folder_id):
        """Convert an ID to a path.

        Parameters
        ----------
        file_or_folder_id : :class:`str`
            The ID of a file or folder.

        Returns
        -------
        :class:`str`
            The corresponding path of the ID.
        """
        names = []
        while True:
            request = self._files.get(
                fileId=file_or_folder_id,
                fields='name,parents',
                supportsAllDrives=True,
            )
            response = request.execute()
            names.append(response['name'])
            parents = response.get('parents', [])
            if not parents:
                break
            if len(parents) > 1:
                raise OSError('Multiple parents exist. This case has not been handled yet. Contact developers.')
            file_or_folder_id = response['parents'][0]
        return '/'.join(names[::-1])

    def move(self, source_id, destination_id):
        """Move a file or a folder.

        When moving a file or folder between `My Drive` and a `Shared drive`
        the access permissions will change.

        Moving a file or folder does not change its ID, only the ID of
        its `parent` changes (i.e., `source_id` will remain the same
        after the move).

        Parameters
        ----------
        source_id : :class:`str`
            The ID of a file or folder to move.
        destination_id : :class:`str`
            The ID of the destination folder. To move the file or folder to the
            `My Drive` root folder then specify ``'root'`` as the `destination_id`.
        """
        params = {'fileId': source_id, 'supportsAllDrives': True}
        try:
            self._files.update(addParents=destination_id, **params).execute()
        except HttpError as e:
            if 'exactly one parent' not in str(e):
                raise

            # Handle the following error:
            #   A shared drive item must have exactly one parent
            response = self._files.get(fields='parents', **params).execute()
            self._files.update(
                addParents=destination_id,
                removeParents=','.join(response['parents']),
                **params).execute()

    def shared_drives(self):
        """Returns the IDs and names of all `Shared drives`.

        Returns
        -------
        :class:`dict`
            The keys are the IDs of the shared drives and the values are the
            names of the shared drives.
        """
        drives = {}
        next_page_token = ''
        while True:
            response = self._drives.list(pageSize=100, pageToken=next_page_token).execute()
            drives.update(dict((d['id'], d['name']) for d in response['drives']))
            next_page_token = response.get('nextPageToken')
            if not next_page_token:
                break
        return drives

    def copy(self, file_id, folder_id=None, name=None):
        """Copy a file.

        Parameters
        ----------
        file_id : :class:`str`
            The ID of a file to copy. Folders cannot be copied.
        folder_id : :class:`str`, optional
            The ID of the destination folder. If not specified then creates
            a copy in the same folder that the original file is located in.
            To copy the file to the `My Drive` root folder then specify
            ``'root'`` as the `folder_id`.
        name : :class:`str`, optional
            The filename of the destination file.

        Returns
        -------
        :class:`str`
            The ID of the destination file.
        """
        response = self._files.copy(
            fileId=file_id,
            fields='id',
            supportsAllDrives=True,
            body={
                'name': name,
                'parents': [folder_id] if folder_id else None,
            },
        ).execute()
        return response['id']

    def rename(self, file_or_folder_id, new_name):
        """Rename a file or folder.

        Renaming a file or folder does not change its ID.

        Parameters
        ----------
        file_or_folder_id : :class:`str`
            The ID of a file or folder.
        new_name : :class:`str`
            The new name of the file or folder.
        """
        self._files.update(
            fileId=file_or_folder_id,
            supportsAllDrives=True,
            body={'name': new_name},
        ).execute()

    def read_only(self, file_id, read_only, reason=''):
        """Set a file to be in read-only mode.

        Parameters
        ----------
        file_id : :class:`str`
            The ID of a file.
        read_only : :class:`bool`
            Whether to set the file to be in read-only mode.
        reason : :class:`str`, optional
            The reason for putting the file in read-only mode.
            Only used if `read_only` is :data:`True`.
        """
        restrictions = {'readOnly': read_only}
        if read_only:
            restrictions['reason'] = reason

            # If `file_id` is already in read-only mode, and it is being set
            # to read-only mode then the API raises a TimeoutError waiting for
            # a response. To avoid this error, check the mode and if it is
            # already in read-only mode we are done.
            if self.is_read_only(file_id):
                return

        self._files.update(
            fileId=file_id,
            supportsAllDrives=True,
            body={'contentRestrictions': [restrictions]}
        ).execute()

    def is_read_only(self, file_id):
        """Returns whether the file is in read-only mode.

        Parameters
        ----------
        file_id : :class:`str`
            The ID of a file.

        Returns
        -------
        :class:`bool`
            Whether the file is in read-only mode.
        """
        response = self._files.get(
            fileId=file_id,
            supportsAllDrives=True,
            fields='contentRestrictions',
        ).execute()
        restrictions = response.get('contentRestrictions')
        if not restrictions:
            return False
        return restrictions[0]['readOnly']


class GValueOption(Enum):
    """Determines how values should be returned."""

    FORMATTED = 'FORMATTED_VALUE'
    """Values will be calculated and formatted in the reply according to the
    cell's formatting. Formatting is based on the spreadsheet's locale, not
    the requesting user's locale. For example, if A1 is 1.23 and A2 is =A1
    and formatted as currency, then A2 would return "$1.23"."""

    UNFORMATTED = 'UNFORMATTED_VALUE'
    """Values will be calculated, but not formatted in the reply.
    For example, if A1 is 1.23 and A2 is =A1 and formatted as currency, then
    A2 would return the number 1.23."""

    FORMULA = 'FORMULA'
    """Values will not be calculated. The reply will include the formulas.
    For example, if A1 is 1.23 and A2 is =A1 and formatted as currency,
    then A2 would return "=A1"."""


class GDateTimeOption(Enum):
    """Determines how dates should be returned."""

    SERIAL_NUMBER = 'SERIAL_NUMBER'
    """Instructs date, time, datetime, and duration fields to be output as
    doubles in "serial number" format, as popularized by Lotus 1-2-3. The
    whole number portion of the value (left of the decimal) counts the days
    since December 30th 1899. The fractional portion (right of the decimal)
    counts the time as a fraction of the day. For example, January 1st 1900
    at noon would be 2.5, 2 because it's 2 days after December 30st 1899,
    and .5 because noon is half a day. February 1st 1900 at 3pm would be
    33.625. This correctly treats the year 1900 as not a leap year."""

    FORMATTED_STRING = 'FORMATTED_STRING'
    """Instructs date, time, datetime, and duration fields to be output as
    strings in their given number format (which is dependent on the
    spreadsheet locale)."""


class GCellType(Enum):
    """The data type of a spreadsheet cell."""

    BOOLEAN = 'BOOLEAN'
    CURRENCY = 'CURRENCY'
    DATE = 'DATE'
    DATE_TIME = 'DATE_TIME'
    EMPTY = 'EMPTY'
    ERROR = 'ERROR'
    NUMBER = 'NUMBER'
    PERCENT = 'PERCENT'
    SCIENTIFIC = 'SCIENTIFIC'
    STRING = 'STRING'
    TIME = 'TIME'
    UNKNOWN = 'UNKNOWN'


GCell = namedtuple('GCell', ('value', 'type', 'formatted'))
"""The information about a Google Sheets cell.

.. attribute:: value
   
   The value of the cell.
   
.. attribute:: type
   
   :class:`GCellType`: The data type of `value`.

.. attribute:: formatted
   
   :class:`str`: The formatted value (i.e., how the `value` is displayed in the cell).
"""


class GSheets(GoogleAPI):

    MIME_TYPE = 'application/vnd.google-apps.spreadsheet'
    SERIAL_NUMBER_ORIGIN = datetime(1899, 12, 30)

    def __init__(self, account=None, credentials=None, read_only=True, scopes=None):
        """Interact with Google Sheets.

        .. attention::
           You must follow the instructions in the prerequisites section for setting up the
           `Sheets API <https://developers.google.com/sheets/api/quickstart/python#prerequisites>`_
           before you can use this class. It is also useful to be aware of the
           `refresh token expiration <https://developers.google.com/identity/protocols/oauth2#expiration>`_
           policy.

        Parameters
        ----------
        account : :class:`str`, optional
            Since a person may have multiple Google accounts, and multiple people
            may run the same code, this parameter decides which token to load
            to authenticate with the Google API. The value can be any text (or
            :data:`None`) that you want to associate with a particular Google
            account, provided that it contains valid characters for a filename.
            The value that you chose when you authenticated with your `credentials`
            should be used for all future instances of this class to access that
            particular Google account. You can associate a different value with
            a Google account at any time (by passing in a different `account`
            value), but you will be asked to authenticate with your `credentials`
            again, or, alternatively, you can rename the token files located in
            :const:`~msl.io.constants.HOME_DIR` to match the new `account` value.
        credentials : :class:`str`, optional
            The path to the `client secrets` OAuth credential file. This
            parameter only needs to be specified the first time that you
            authenticate with a particular Google account or if you delete
            the token file that was created when you previously authenticated.
        read_only : :class:`bool`, optional
            Whether to interact with Google Sheets in read-only mode.
        scopes : :class:`list` of :class:`str`, optional
            The list of scopes to enable for the Google API. See
            `Sheets scopes <https://developers.google.com/identity/protocols/oauth2/scopes#sheets>`_
            for more details. If not specified then default scopes are chosen
            based on the value of `read_only`.
        """
        if not scopes:
            if read_only:
                scopes = ['https://www.googleapis.com/auth/spreadsheets.readonly']
            else:
                scopes = ['https://www.googleapis.com/auth/spreadsheets']

        super(GSheets, self).__init__(
            'sheets', 'v4', credentials, scopes, read_only, account)

        self._spreadsheets = self._service.spreadsheets()

    def append(self, values, spreadsheet_id, cell=None, sheet=None, row_major=True, raw=False):
        """Append values to a sheet.

        Returns
        -------
        values
            The value(s) to append
        spreadsheet_id : :class:`str`
            The ID of a Google Sheets file.
        cell : :class:`str`, optional
            The cell (top-left corner) to start appending the values to. If the
            cell already contains data then new rows are inserted and the values
            are written to the new rows. For example, ``'D100'``.
        sheet : :class:`str`, optional
            The name of a sheet in the spreadsheet to append the values to.
            If not specified and only one sheet exists in the spreadsheet
            then automatically determines the sheet name; however, it is
            more efficient to specify the name of the sheet.
        row_major : :class:`bool`, optional
            Whether to append the values in row-major or column-major order.
        raw : :class:`bool`, optional
            Determines how the values should be interpreted. If :data:`True`,
            the values will not be parsed and will be stored as-is. If
            :data:`False`, the values will be parsed as if the user typed
            them into the UI. Numbers will stay as numbers, but strings may
            be converted to numbers, dates, etc. following the same rules
            that are applied when entering text into a cell via the Google
            Sheets UI.
        """
        self._spreadsheets.values().append(
            spreadsheetId=spreadsheet_id,
            range=self._get_range(sheet, cell, spreadsheet_id),
            valueInputOption='RAW' if raw else 'USER_ENTERED',
            insertDataOption='INSERT_ROWS',
            body={
                'values': self._values(values),
                'majorDimension': 'ROWS' if row_major else 'COLUMNS',
            },
        ).execute()

    def write(self, values, spreadsheet_id, cell, sheet=None, row_major=True, raw=False):
        """Write values to a sheet.

        If a cell that is being written to already contains a value,
        the value in that cell is overwritten with the new value.

        Returns
        -------
        values
            The value(s) to write.
        spreadsheet_id : :class:`str`
            The ID of a Google Sheets file.
        cell : :class:`str`, optional
            The cell (top-left corner) to start writing the values to.
            For example, ``'C9'``.
        sheet : :class:`str`, optional
            The name of a sheet in the spreadsheet to write the values to.
            If not specified and only one sheet exists in the spreadsheet
            then automatically determines the sheet name; however, it is
            more efficient to specify the name of the sheet.
        row_major : :class:`bool`, optional
            Whether to write the values in row-major or column-major order.
        raw : :class:`bool`, optional
            Determines how the values should be interpreted. If :data:`True`,
            the values will not be parsed and will be stored as-is. If
            :data:`False`, the values will be parsed as if the user typed
            them into the UI. Numbers will stay as numbers, but strings may
            be converted to numbers, dates, etc. following the same rules
            that are applied when entering text into a cell via the Google
            Sheets UI.
        """
        self._spreadsheets.values().update(
            spreadsheetId=spreadsheet_id,
            range=self._get_range(sheet, cell, spreadsheet_id),
            valueInputOption='RAW' if raw else 'USER_ENTERED',
            body={
                'values': self._values(values),
                'majorDimension': 'ROWS' if row_major else 'COLUMNS',
            },
        ).execute()

    def copy(self, name_or_id, spreadsheet_id, destination_spreadsheet_id):
        """Copy a sheet from one spreadsheet to another spreadsheet.

        Parameters
        ----------
        name_or_id : :class:`str` or :class:`int`
            The name or ID of the sheet to copy.
        spreadsheet_id : :class:`str`
            The ID of the spreadsheet that contains the sheet.
        destination_spreadsheet_id : :class:`str`
            The ID of a spreadsheet to copy the sheet to.

        Returns
        -------
        :class:`int`
            The ID of the sheet in the destination spreadsheet.
        """
        if isinstance(name_or_id, int):
            sheet_id = name_or_id
        else:
            sheet_id = self.sheet_id(name_or_id, spreadsheet_id)

        response = self._spreadsheets.sheets().copyTo(
            spreadsheetId=spreadsheet_id,
            sheetId=sheet_id,
            body={
                'destination_spreadsheet_id': destination_spreadsheet_id,
            },
        ).execute()
        return response['sheetId']

    def sheet_id(self, name, spreadsheet_id):
        """Returns the ID of a sheet.

        Parameters
        ----------
        name : :class:`str`
            The name of the sheet.
        spreadsheet_id : :class:`str`
            The ID of the spreadsheet.

        Returns
        -------
        :class:`int`
            The ID of the sheet.
        """
        request = self._spreadsheets.get(spreadsheetId=spreadsheet_id)
        response = request.execute()
        for sheet in response['sheets']:
            if sheet['properties']['title'] == name:
                return sheet['properties']['sheetId']
        raise ValueError('There is no sheet named {!r}'.format(name))

    def rename_sheet(self, name_or_id, new_name, spreadsheet_id):
        """Rename a sheet.

        Parameters
        ----------
        name_or_id : :class:`str` or :class:`int`
            The name or ID of the sheet to rename.
        new_name : :class:`str`
            The new name of the sheet.
        spreadsheet_id : :class:`str`
            The ID of the spreadsheet that contains the sheet.
        """
        if isinstance(name_or_id, int):
            sheet_id = name_or_id
        else:
            sheet_id = self.sheet_id(name_or_id, spreadsheet_id)

        self._spreadsheets.batchUpdate(
            spreadsheetId=spreadsheet_id,
            body={
                'requests': [{
                    'updateSheetProperties': {
                        'properties': {
                            'sheetId': sheet_id,
                            'title': new_name,
                        },
                        'fields': 'title',
                    }
                }]
            }
        ).execute()

    def add_sheets(self, names, spreadsheet_id):
        """Add sheets to a spreadsheet.

        Parameters
        ----------
        names : :class:`str` or :class:`list` of :class:`str`
            The name(s) of the new sheet(s) to add.
        spreadsheet_id : :class:`str`
            The ID of the spreadsheet to add the sheet(s) to.

        Returns
        -------
        :class:`dict`
            The keys are the IDs of the new sheets and the values are the names.
        """
        if isinstance(names, str):
            names = [names]
        response = self._spreadsheets.batchUpdate(
            spreadsheetId=spreadsheet_id,
            body={
                'requests': [{
                    'addSheet': {
                        'properties': {
                            'title': name
                        }
                    }
                } for name in names]
            }
        ).execute()
        return OrderedDict((r['addSheet']['properties']['sheetId'],
                            r['addSheet']['properties']['title'])
                           for r in response['replies'])

    def delete_sheets(self, names_or_ids, spreadsheet_id):
        """Delete sheets from a spreadsheet.

        Parameters
        ----------
        names_or_ids : :class:`str`, :class:`int` or :class:`list`
            The name(s) or ID(s) of the sheet(s) to delete.
        spreadsheet_id : :class:`str`
            The ID of the spreadsheet to delete the sheet(s) from.
        """
        if not isinstance(names_or_ids, (list, tuple)):
            names_or_ids = [names_or_ids]
        self._spreadsheets.batchUpdate(
            spreadsheetId=spreadsheet_id,
            body={
                'requests': [{
                    'deleteSheet': {
                        'sheetId': n if isinstance(n, int) else self.sheet_id(n, spreadsheet_id)
                    }
                } for n in names_or_ids]
            }
        ).execute()

    def create(self, name, sheet_names=None):
        """Create a new spreadsheet.

        The spreadsheet will be created in the `My Drive` root folder.
        To move it to a different folder use :meth:`GDrive.create_folder`
        and/or :meth:`GDrive.move`.

        Parameters
        ----------
        name : :class:`str`
            The name of the spreadsheet.
        sheet_names : :class:`list` of :class:`str`, optional
            The names of the sheets that are in the spreadsheet.

        Returns
        -------
        :class:`str`
            The ID of the spreadsheet that was created.
        """
        body = {'properties': {'title': name}}
        if sheet_names:
            body['sheets'] = [{
                'properties': {'title': sn}
            } for sn in sheet_names]
        response = self._spreadsheets.create(body=body).execute()
        return response['spreadsheetId']

    def sheet_names(self, spreadsheet_id):
        """Get the names of all sheets in a spreadsheet.

        Parameters
        ----------
        spreadsheet_id : :class:`str`
            The ID of a Google Sheets file.

        Returns
        -------
        :class:`tuple` of :class:`str`
            The names of all sheets.
        """
        request = self._spreadsheets.get(spreadsheetId=spreadsheet_id)
        response = request.execute()
        return tuple(r['properties']['title'] for r in response['sheets'])

    def values(self,
               spreadsheet_id,
               sheet=None,
               cells=None,
               row_major=True,
               value_option=GValueOption.FORMATTED,
               datetime_option=GDateTimeOption.SERIAL_NUMBER
               ):
        """Return a range of values from a spreadsheet.

        Parameters
        ----------
        spreadsheet_id : :class:`str`
            The ID of a Google Sheets file.
        sheet : :class:`str`, optional
            The name of a sheet in the spreadsheet to read the values from.
            If not specified and only one sheet exists in the spreadsheet
            then automatically determines the sheet name; however, it is
            more efficient to specify the name of the sheet.
        cells : :class:`str`, optional
            The A1 notation or R1C1 notation of the range to retrieve values
            from. If not specified then returns all values that are in `sheet`.
        row_major : :class:`bool`, optional
            Whether to return the values in row-major or column-major order.
        value_option : :class:`str` or :class:`GValueOption`, optional
            How values should be represented in the output. If a string
            then it must be equal to one of the values in :class:`GValueOption`.
        datetime_option : :class:`str` or :class:`GDateTimeOption`, optional
            How dates, times, and durations should be represented in the
            output. If a string then it must be equal to one of the values
            in :class:`GDateTimeOption`. This argument is ignored if
            `value_option` is :attr:`GValueOption.FORMATTED`.

        Returns
        -------
        :class:`list`
            The values from the sheet.
        """
        if hasattr(value_option, 'value'):
            value_option = value_option.value

        if hasattr(datetime_option, 'value'):
            datetime_option = datetime_option.value

        response = self._spreadsheets.values().get(
            spreadsheetId=spreadsheet_id,
            range=self._get_range(sheet, cells, spreadsheet_id),
            majorDimension='ROWS' if row_major else 'COLUMNS',
            valueRenderOption=value_option,
            dateTimeRenderOption=datetime_option
        ).execute()
        return response.get('values', [])

    def cells(self, spreadsheet_id, ranges=None):
        """Return cells from a spreadsheet.

        Parameters
        ----------
        spreadsheet_id : :class:`str`
            The ID of a Google Sheets file.
        ranges : :class:`str` or :class:`list` of :class:`str`, optional
            The ranges to retrieve from the spreadsheet. Examples:

                * ``'Sheet1'`` :math:`\\rightarrow` return all cells from
                  the sheet named Sheet1
                * ``'Sheet1!A1:H5'`` :math:`\\rightarrow` return cells A1:H5
                  from the sheet named Sheet1
                * ``['Sheet1!A1:H5', 'Data', 'Devices!B4:B9']`` :math:`\\rightarrow`
                  return cells A1:H5 from the sheet named Sheet1, all cells from the
                  sheet named Data and cells B4:B9 from the sheet named Devices

            If not specified then return all cells from all sheets.

        Returns
        -------
        :class:`dict`
            The cells from the spreadsheet. The keys are the names of the
            sheets and the values are a :class:`list` of :class:`GCell`
            objects for the specified range of each sheet.
        """
        response = self._spreadsheets.get(
            spreadsheetId=spreadsheet_id,
            includeGridData=True,
            ranges=ranges,
        ).execute()
        cells = {}
        for sheet in response['sheets']:
            data = []
            for item in sheet['data']:
                for row in item.get('rowData', []):
                    row_data = []
                    for col in row.get('values', []):
                        effective_value = col.get('effectiveValue', None)
                        formatted = col.get('formattedValue', '')
                        if effective_value is None:
                            value = None
                            typ = GCellType.EMPTY
                        elif 'numberValue' in effective_value:
                            value = effective_value['numberValue']
                            t = col.get('effectiveFormat', {}).get('numberFormat', {}).get('type', 'NUMBER')
                            typ = GCellType(t)
                        elif 'stringValue' in effective_value:
                            value = effective_value['stringValue']
                            typ = GCellType.STRING
                        elif 'boolValue' in effective_value:
                            value = effective_value['boolValue']
                            typ = GCellType.BOOLEAN
                        elif 'errorValue' in effective_value:
                            msg = effective_value['errorValue']['message']
                            value = '{} ({})'.format(col['formattedValue'], msg)
                            typ = GCellType.ERROR
                        else:
                            value = formatted
                            typ = GCellType.UNKNOWN
                        row_data.append(GCell(value=value, type=typ, formatted=formatted))
                    data.append(row_data)
                cells[sheet['properties']['title']] = data
        return cells

    @staticmethod
    def to_datetime(value):
        """Convert a "serial number" date into a :class:`datetime.datetime`.

        Parameters
        ----------
        value : :class:`float`
            A date in the "serial number" format.

        Returns
        -------
        :class:`datetime.datetime`
            The date converted.
        """
        days = int(value)
        seconds = (value - days) * 86400  # 60 * 60 * 24
        return GSheets.SERIAL_NUMBER_ORIGIN + timedelta(days=days, seconds=seconds)

    def _get_range(self, sheet, cells, spreadsheet_id):
        if not sheet:
            names = self.sheet_names(spreadsheet_id)
            if len(names) != 1:
                sheets = ', '.join(repr(n) for n in names)
                raise ValueError('You must specify a sheet name: ' + sheets)
            _range = names[0]
        else:
            _range = sheet

        if cells:
            _range += '!{}'.format(cells)

        return _range

    @staticmethod
    def _values(values):
        """The append() and update() API methods require a list of lists."""
        if not isinstance(values, (list, tuple)):
            return [[values]]
        if values and not isinstance(values[0], (list, tuple)):
            return [values]
        return values


class GMail(GoogleAPI):

    def __init__(self, account=None, credentials=None, scopes=None):
        """Interact with Gmail.

        .. attention::
           You must follow the instructions in the prerequisites section for setting up the
           `Gmail API <https://developers.google.com/gmail/api/quickstart/python#prerequisites>`_
           before you can use this class. It is also useful to be aware of the
           `refresh token expiration <https://developers.google.com/identity/protocols/oauth2#expiration>`_
           policy.

        Parameters
        ----------
        account : :class:`str`, optional
            Since a person may have multiple Google accounts, and multiple people
            may run the same code, this parameter decides which token to load
            to authenticate with the Google API. The value can be any text (or
            :data:`None`) that you want to associate with a particular Google
            account, provided that it contains valid characters for a filename.
            The value that you chose when you authenticated with your `credentials`
            should be used for all future instances of this class to access that
            particular Google account. You can associate a different value with
            a Google account at any time (by passing in a different `account`
            value), but you will be asked to authenticate with your `credentials`
            again, or, alternatively, you can rename the token files located in
            :const:`~msl.io.constants.HOME_DIR` to match the new `account` value.
        credentials : :class:`str`, optional
            The path to the `client secrets` OAuth credential file. This
            parameter only needs to be specified the first time that you
            authenticate with a particular Google account or if you delete
            the token file that was created when you previously authenticated.
        scopes : :class:`list` of :class:`str`, optional
            The list of scopes to enable for the Google API. See
            `Gmail scopes <https://developers.google.com/identity/protocols/oauth2/scopes#gmail>`_
            for more details. If not specified then default scopes are chosen.
        """
        if not scopes:
            scopes = [
                'https://www.googleapis.com/auth/gmail.send',
                'https://www.googleapis.com/auth/gmail.metadata'
            ]

        super(GMail, self).__init__(
            'gmail', 'v1', credentials, scopes, False, account)

        self._my_email_address = None
        self._users = self._service.users()

    def profile(self):
        """Gets the authenticated user's Gmail profile.

        Returns
        -------
        :class:`dict`
            Returns the following

            .. code-block:: console

                {
                   'email_address': string, The authenticated user's email address
                   'messages_total': integer, The total number of messages in the mailbox
                   'threads_total': integer, The total number of threads in the mailbox
                   'history_id': string, The ID of the mailbox's current history record
                }

        """
        profile = self._users.getProfile(userId='me').execute()
        return {
            'email_address': profile['emailAddress'],
            'messages_total': profile['messagesTotal'],
            'threads_total': profile['threadsTotal'],
            'history_id': profile['historyId'],
        }

    def send(self, recipients, sender='me', subject=None, body=None):
        """Send an email.

        Parameters
        ----------
        recipients : :class:`str` or :class:`list` of :class:`str`
            The email address(es) of the recipient(s). The value ``'me'``
            can be used to indicate the authenticated user.
        sender : :class:`str`, optional
            The email address of the sender. The value ``'me'``
            can be used to indicate the authenticated user.
        subject : :class:`str`, optional
            The text to include in the subject field.
        body : :class:`str`, optional
            The text to include in the body of the email. The text can be
            enclosed in ``<html></html>`` tags to use HTML elements to format
            the message.

        See Also
        --------
        :func:`~msl.io.utils.send_email`
        """
        if isinstance(recipients, str):
            recipients = [recipients]

        for i in range(len(recipients)):
            if recipients[i] == 'me':
                if self._my_email_address is None:
                    self._my_email_address = self.profile()['email_address']
                recipients[i] = self._my_email_address

        msg = MIMEMultipart()
        msg['From'] = sender
        msg['To'] = ', '.join(recipients)
        msg['Subject'] = subject or '(no subject)'

        text = body or ''
        subtype = 'html' if text.startswith('<html>') else 'plain'
        msg.attach(MIMEText(text, subtype))

        self._users.messages().send(
            userId=sender,
            body={'raw': b64encode(msg.as_bytes()).decode()}
        ).execute()



from __future__ import unicode_literals

import os
import pathlib
import sys
from datetime import datetime
from io import BytesIO
from io import StringIO

import numpy as np
import pytest

from helper import datasets_equal
from msl.io import read_table
from msl.io.constants import IS_PYTHON2
from msl.io.tables import read_table_excel
from msl.io.tables import read_table_gsheets
from test_google_api import skipif_no_gdrive_readonly
from test_google_api import skipif_no_sheets_readonly

skipif_32bit_py27 = pytest.mark.skipif(
    sys.maxsize < 2**32 and IS_PYTHON2,
    reason='Avoid Google API MemoryError on 32-bit Python 2.7'
)

# the data in the Excel, CVS and TXT files that are tested contain the following
header = np.asarray(['timestamp', 'val1', 'uncert1', 'val2', 'uncert2'], dtype=str)
data = np.asarray([
    ('2019-09-11 14:06:55', -0.505382, 0.000077, 0.501073, 0.000079),
    ('2019-09-11 14:06:59', -0.505191, 0.000066, 0.500877, 0.000083),
    ('2019-09-11 14:07:03', -0.505308, 0.000086, 0.500988, 0.000087),
    ('2019-09-11 14:07:07', -0.505250, 0.000119, 0.500923, 0.000120),
    ('2019-09-11 14:07:11', -0.505275, 0.000070, 0.500965, 0.000088),
    ('2019-09-11 14:07:15', -0.505137, 0.000079, 0.500817, 0.000085),
    ('2019-09-11 14:07:19', -0.505073, 0.000099, 0.500786, 0.000084),
    ('2019-09-11 14:07:23', -0.505133, 0.000088, 0.500805, 0.000076),
    ('2019-09-11 14:07:27', -0.505096, 0.000062, 0.500759, 0.000062),
    ('2019-09-11 14:07:31', -0.505072, 0.000142, 0.500739, 0.000149),
], dtype='U19,f8,f8,f8,f8')

# the data in the GSheet spreadsheet
gsheet_header = np.asarray(['Timestamp', 'Value', 'Valid', 'ID'])
gsheet_data = np.asarray([
    (datetime(2019, 9, 11, 14, 6, 55), 20.1, True, 'sensor 1'),
    (datetime(2019, 9, 11, 14, 6, 59), 25.4, False, 'sensor 2'),
    (datetime(2019, 9, 11, 14, 7, 3), 19.4, True, 'sensor 3'),
    (datetime(2019, 9, 11, 14, 7, 7), 11.8, False, 'sensor 4'),
    (datetime(2019, 9, 11, 14, 7, 11), 24.6, False, 'sensor 5'),
    (datetime(2019, 9, 11, 14, 7, 15), 20.7, True, 'sensor 1'),
    (datetime(2019, 9, 11, 14, 7, 19), 21.8, True, 'sensor 2'),
    (datetime(2019, 9, 11, 14, 7, 23), 19.2, True, 'sensor 3'),
    (datetime(2019, 9, 11, 14, 7, 27), 18.6, False, 'sensor 4'),
    (datetime(2019, 9, 11, 14, 7, 31), 16.4, False, 'sensor 5'),
])


INVALID_CELL_RANGES = ['', ':', 'A', 'A:', '1', 'ZZ', 'AB', 'A-B', 'A1D10', 'A1-D10']


def get_url(extension):
    return os.path.join(os.path.dirname(__file__), 'samples', 'table' + extension)


def test_raises():

    # file does not exist
    with pytest.raises((IOError, OSError)):
        read_table('does not exist')

    # the 'unpack' argument is not supported for text-based files
    with pytest.raises(ValueError, match='unpack'):
        read_table(get_url('.csv'), unpack=True)

    # invalid cell range
    for c in INVALID_CELL_RANGES:
        with pytest.raises(ValueError, match=r'Invalid cell'):
            read_table(get_url('.xls'), cells=c, sheet='A1')


def test_excel_range_out_of_bounds():
    for c in ['A100', 'J1:M10']:
        dset = read_table(get_url('.xls'), cells=c, sheet='A1')
        assert dset.metadata.header.size == 0
        assert dset.size == 0

    dset = read_table(get_url('.xls'), cells='A1:Z100', sheet='A1', as_datetime=False, dtype=data.dtype)
    assert np.array_equal(dset.metadata.header, header)
    assert np.array_equal(dset.data, data)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(dtype=data.dtype)),
     ('.txt', dict(dtype=data.dtype, delimiter='\t')),
     ('.xls', dict(dtype=data.dtype, sheet='A1', as_datetime=False)),
     ('.xls', dict(dtype=data.dtype, sheet='BH11', as_datetime=False, cells='BH11:BL21')),
     ('.xlsx', dict(dtype=data.dtype, sheet='A1', as_datetime=False)),
     ('.xlsx', dict(dtype=data.dtype, sheet='BH11', as_datetime=False, cells='BH11:BL21')),
     ('.xlsx', dict(dtype=data.dtype, sheet='AEX154041', as_datetime=False, cells='AEX154041:AFB154051'))]
)
def test_fetch_all_data(extn, kwargs):
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, header)
    assert np.array_equal(dset.data, data)
    assert dset.shape == (10,)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(usecols=(1, 2, 3, 4))),
     ('.txt', dict(usecols=(1, 2, 3, 4), delimiter='\t')),
     ('.xls', dict(sheet='A1', cells='B1:E11')),
     ('.xls', dict(sheet='BH11', cells='BI11:BL21')),
     ('.xlsx', dict(sheet='A1', cells='B1:E11')),
     ('.xlsx', dict(sheet='BH11', cells='BI11:BL21')),
     ('.xlsx', dict(sheet='AEX154041', cells='AEY154041:AFB154051'))]
)
def test_ignore_timestamp_column(extn, kwargs):
    floats = np.asarray([[e['f1'], e['f2'], e['f3'], e['f4']] for e in data])
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, header[1:])
    assert np.array_equal(dset.data, floats)
    assert dset.shape == (10, 4)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(usecols=1)),
     ('.txt', dict(usecols=1, delimiter='\t')),
     ('.xls', dict(sheet='A1', cells='B1:B11')),
     ('.xls', dict(sheet='BH11', cells='BI11:BI21')),
     ('.xlsx', dict(sheet='A1', cells='B1:B11')),
     ('.xlsx', dict(sheet='BH11', cells='BI11:BI21')),
     ('.xlsx', dict(sheet='AEX154041', cells='AEY154041:AEY154051'))]
)
def test_single_column(extn, kwargs):
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, [header[1]])
    assert np.array_equal(dset.data, data['f1'])
    assert dset.shape == (10,)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(dtype=data.dtype, max_rows=1)),
     ('.txt', dict(dtype=data.dtype, max_rows=1, delimiter='\t')),
     ('.xls', dict(dtype=data.dtype, sheet='A1', as_datetime=False, cells='A1:E2')),
     ('.xls', dict(dtype=data.dtype, sheet='BH11', as_datetime=False, cells='BH11:BL12')),
     ('.xlsx', dict(dtype=data.dtype, sheet='A1', as_datetime=False, cells='A1:E2')),
     ('.xlsx', dict(dtype=data.dtype, sheet='BH11', as_datetime=False, cells='BH11:BL12')),
     ('.xlsx', dict(dtype=data.dtype, sheet='AEX154041', as_datetime=False, cells='AEX154041:AFB154042'))]
)
def test_single_row(extn, kwargs):
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, header)
    assert np.array_equal(dset.data, data[0])


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.xls', dict(sheet='A1', cells='A1:E1')),
     ('.xls', dict(sheet='BH11', cells='BH11:BL11')),
     ('.xlsx', dict(sheet='A1', cells='A1:E1')),
     ('.xlsx', dict(sheet='BH11', cells='BH11:BL11')),
     ('.xlsx', dict(sheet='AEX154041', cells='AEX154041:AFB154041'))]
)
def test_header_only(extn, kwargs):
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, header)
    assert dset.data.size == 0


def to_datetime(string):
    return datetime.strptime(string.decode(), '%Y-%m-%d %H:%M:%S')


dt = {'names': header, 'formats': [object, float, float, float, float]}


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(dtype=dt, converters={0: to_datetime})),
     ('.txt', dict(dtype=dt, converters={0: to_datetime}, delimiter='\t')),
     ('.xls', dict(dtype=dt, sheet='A1')),
     ('.xls', dict(dtype=dt, sheet='BH11', cells='BH11:BL21')),
     ('.xlsx', dict(dtype=dt, sheet='A1')),
     ('.xlsx', dict(dtype=dt, sheet='BH11', cells='BH11:BL21')),
     ('.xlsx', dict(dtype=dt, sheet='AEX154041', cells='AEX154041:AFB154051'))]
)
def test_datetime_objects(extn, kwargs):
    datetimes = np.asarray([to_datetime(item.encode()) for item in data['f0']], dtype=object)
    data_datetimes = np.asarray([
        (a, b, c, d, e) for a, b, c, d, e in zip(datetimes, data['f1'], data['f2'], data['f3'], data['f4'])
    ], dtype=dt)

    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, header)
    for h in header:
        assert np.array_equal(dset[h], data_datetimes[h])
    assert dset.shape == (10,)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(dtype=data.dtype, skiprows=5)),
     ('.txt', dict(dtype=data.dtype, delimiter='\t', skiprows=5))]
)
def test_skip_rows_5(extn, kwargs):
    new_header = ['2019-09-11 14:07:11', '-0.505275', '0.000070', '0.500965', '0.000088']
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, new_header)
    assert np.array_equal(dset.data, data[5:])
    assert dset.shape == (5,)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(skiprows=3, usecols=(1, 4))),
     ('.txt', dict(skiprows=3, usecols=(1, 4), delimiter='\t'))]
)
def test_skip_rows_use_cols(extn, kwargs):
    h = ['-0.505308', '0.000087']
    d = [[-0.505250, 0.000120],
         [-0.505275, 0.000088],
         [-0.505137, 0.000085],
         [-0.505073, 0.000084],
         [-0.505133, 0.000076],
         [-0.505096, 0.000062],
         [-0.505072, 0.000149]]
    dset = read_table(get_url(extn), **kwargs)
    assert np.array_equal(dset.metadata.header, h)
    assert np.array_equal(dset.data, d)
    assert dset.shape == (7, 2)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(skiprows=100)),
     ('.txt', dict(skiprows=100))]
)
def test_skip_rows_100(extn, kwargs):
    dset = read_table(get_url(extn), **kwargs)
    assert dset.metadata.header.size == 0
    assert dset.size == 0


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.csv', dict(dtype=data.dtype, delimiter=',')),
     ('.txt', dict(dtype=data.dtype, delimiter='\t'))]
)
def test_text_file_like(extn, kwargs):
    def assert_dataset(dataset):
        assert np.array_equal(dataset.metadata.header, header)
        assert np.array_equal(dataset.data, data)
        assert dataset.shape == (10,)

    # first, load it using the file path to get it into a Dataset object
    dset_temp = read_table(get_url(extn), **kwargs)

    with StringIO() as buf:
        delim = kwargs['delimiter']
        buf.write(delim.join(h for h in dset_temp.metadata.header) + '\n')
        for row in dset_temp:
            buf.write(delim.join(str(val) for val in row) + '\n')

        buf.seek(0)
        dset = read_table(buf, **kwargs)
        assert dset.name == 'StringIO'
        assert_dataset(dset)

    with open(get_url(extn), mode='rt') as fp:
        dset = read_table(fp, **kwargs)
        assert dset.name == 'table' + extn
        assert_dataset(dset)

    kwargs['delimiter'] = kwargs['delimiter'].encode()

    with BytesIO() as buf:
        delim = kwargs['delimiter']
        buf.write(delim.join(h.encode() for h in dset_temp.metadata.header) + b'\n')
        for row in dset_temp:
            buf.write(delim.join(str(val).encode() for val in row) + b'\n')

        buf.seek(0)
        dset = read_table(buf, **kwargs)
        assert dset.name == 'BytesIO'
        assert_dataset(dset)

    with open(get_url(extn), mode='rb') as fp:
        dset = read_table(fp, **kwargs)
        assert dset.name == 'table' + extn
        assert_dataset(dset)


@pytest.mark.parametrize(
    ('extn', 'kwargs'),
    [('.xls', dict(dtype=data.dtype, sheet='A1', as_datetime=False)),
     ('.xlsx', dict(dtype=data.dtype, sheet='A1', as_datetime=False))]
)
def test_excel_file_pointer(extn, kwargs):
    for mode in ['rt', 'rb']:
        with open(get_url(extn), mode=mode) as fp:
            dataset = read_table(fp, **kwargs)
            assert np.array_equal(dataset.metadata.header, header)
            assert np.array_equal(dataset.data, data)
            assert dataset.shape == (10,)
        with open(get_url(extn), mode=mode) as fp:
            dataset = read_table_excel(fp, **kwargs)
            assert np.array_equal(dataset.metadata.header, header)
            assert np.array_equal(dataset.data, data)
            assert dataset.shape == (10,)

    # there is no point to test StringIO nor BytesIO because `read_table`
    # checks the file path extension to decide how to read the table
    # and a StringIO and a BytesIO object do not have an extension to check
    # so read_table_excel will not be called, also xlrd cannot load a file stream


def test_pathlib():
    string = os.path.join(os.path.dirname(__file__), 'samples', 'table.csv')
    dset1 = read_table(string, dtype=object)
    dset2 = read_table(pathlib.Path(string), dtype=object)
    assert datasets_equal(dset1, dset2)

    string = os.path.join(os.path.dirname(__file__), 'samples', 'table.xls')
    dset1 = read_table(string, sheet='A1')
    dset2 = read_table(pathlib.Path(string), sheet='A1')
    assert datasets_equal(dset1, dset2)


@skipif_32bit_py27
@skipif_no_gdrive_readonly
@skipif_no_sheets_readonly
def test_gsheet_file_path():
    dset = read_table('table.gsheet', account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table('MSL/msl-io-testing/Copy of table.gsheet', account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)


@skipif_32bit_py27
@skipif_no_gdrive_readonly
@skipif_no_sheets_readonly
def test_gsheet_pathlib():
    dset = read_table(pathlib.Path('table.gsheet'), account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    path = pathlib.Path('MSL/msl-io-testing/Copy of table.gsheet')
    dset = read_table(path, account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table_gsheets(path, account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)


@skipif_32bit_py27
@skipif_no_gdrive_readonly
@skipif_no_sheets_readonly
def test_gsheet_file_pointer():
    filename = 'table.gsheet'
    with open(filename, mode='w'):
        pass

    for m in ['rt', 'rb']:
        with open(filename, mode=m) as fp:
            dset = read_table(fp, account='testing', sheet='StartA1')
        assert np.array_equal(dset.metadata.header, gsheet_header)
        assert np.array_equal(dset, gsheet_data)

        with open(filename, mode=m) as fp:
            dset = read_table_gsheets(fp, account='testing', sheet='StartA1')
        assert np.array_equal(dset.metadata.header, gsheet_header)
        assert np.array_equal(dset, gsheet_data)

    os.remove(filename)

    # there is no point to test StringIO nor BytesIO because `read_table`
    # checks the file path extension to decide how to read the table
    # and a StringIO and a BytesIO object do not have an extension to check
    # so read_table_gsheets will not be called, also GSheets cannot load a file stream


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_as_datetime():
    # ID of the table.gsheet file
    table_id = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'

    dset = read_table(table_id, account='testing', sheet='StartA1', as_datetime=False)
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset.data, gsheet_data.astype(str))

    dset = read_table(table_id, account='testing', sheet='StartA1', as_datetime=False, dtype=object)
    data2 = gsheet_data.copy()
    data2[:, 0] = [str(item) for item in gsheet_data[:, 0]]
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, data2)


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_all_data():
    # ID of the table.gsheet file
    table_id = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ'

    dset = read_table(table_id+'.gsheet', account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table_gsheets(table_id, account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table(table_id+'.gsheet', account='testing', sheet='StartH22')
    assert dset.metadata.header.size == 11
    assert dset.shape == (31, 11)
    for i in range(20):
        assert np.array_equal(dset[i], [None] * 11)
    assert np.array_equal(dset[20][:7], [None] * 7)
    assert np.array_equal(dset[20][7:], gsheet_header)
    for i, row in enumerate(gsheet_data, start=21):
        assert np.array_equal(dset[i][:7], [None] * 7)
        assert np.array_equal(dset[i][7:], row)

    dset = read_table(table_id+'.gsheet', account='testing', sheet='StartH22', cells='H22')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    # ID of MSL/msl-io-testing/Copy of table.gsheet
    file = '1NfDUZzHk71CPAfhIoE8l9h4NJ8oeqKfqGAUM81Vyc88.gsheet'
    dset = read_table(file, account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_one_row():
    # ID of the table.gsheet file
    file = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'
    dset = read_table(file, account='testing', sheet='row')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data[0])


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_one_column():
    # ID of the table.gsheet file
    file = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'
    dset = read_table(file, account='testing', sheet='column')
    assert np.array_equal(dset.metadata.header, [gsheet_header[1]])
    assert np.array_equal(dset, gsheet_data[:, 1])


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_header_only():
    # ID of the table.gsheet file
    file = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'
    dset = read_table(file, account='testing', sheet='header only')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert dset.size == 0


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_empty():
    # ID of the table.gsheet file
    file = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'
    dset = read_table(file, account='testing', sheet='empty')
    assert dset.metadata.header.size == 0
    assert dset.size == 0


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheets_cell_range():
    # ID of the table.gsheet file
    file = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'

    dset = read_table(file, account='testing', sheet='StartH22', cells='H22')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table(file, account='testing', sheet='StartH22', cells='H22:K32')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table(file, account='testing', sheet='StartA1', cells='A1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data)

    dset = read_table(file, account='testing', sheet='StartA1', cells='A2')
    assert np.array_equal(dset.metadata.header, [str(item) for item in gsheet_data[0]])
    assert np.array_equal(dset, gsheet_data[1:])

    dset = read_table(file, account='testing', sheet='StartA1', cells='A4:C7')
    assert np.array_equal(dset.metadata.header, ['2019-09-11 14:07:03', '19.4', 'True'])
    assert np.array_equal(dset, gsheet_data[3:6, :3])

    dset = read_table(file, account='testing', sheet='StartA1', cells='B1:B11')
    assert np.array_equal(dset.metadata.header, [gsheet_header[1]])
    assert np.array_equal(dset, gsheet_data[:, 1])

    dset = read_table(file, account='testing', sheet='StartA1', cells='A1:D2')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset, gsheet_data[0])


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheet_range_out_of_bounds():
    for c in ['A100', 'J1:M10']:
        dset = read_table('1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet',
                          cells=c, account='testing', sheet='StartA1')
        assert dset.metadata.header.size == 0
        assert dset.size == 0

    dset = read_table('1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet',
                      cells='A1:Z100', account='testing', sheet='StartA1')
    assert np.array_equal(dset.metadata.header, gsheet_header)
    assert np.array_equal(dset.data, gsheet_data)


@skipif_32bit_py27
@skipif_no_sheets_readonly
def test_gsheet_raises():

    ssid = '1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ.gsheet'

    # invalid cell range
    for c in INVALID_CELL_RANGES:
        with pytest.raises(ValueError, match=r'Invalid cell'):
            read_table(ssid, cells=c, sheet='StartA1', read_only=True, account='testing')



import datetime
import hashlib
import os
import re
import shutil
import stat
import subprocess
import sys
import tempfile
import uuid
from io import BytesIO
from io import StringIO

import pytest

from msl.io import utils

try:
    PermissionError
except NameError:
    PermissionError = OSError  # for Python 2.7


def test_search():

    def s(**kwargs):
        return list(utils.search(base, **kwargs))

    # the msl-io folder
    base = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
    assert base.endswith('msl-io')

    # (?!c) means match __init__.py files but ignore __init__.pyc files
    files = s(pattern=r'__init__\.py(?!c)')
    assert len(files) == 0

    files = s(pattern=r'__init__\.py(?!c)', levels=1)
    assert len(files) == 1

    files = s(pattern=r'__init__\.py(?!c)', levels=2)
    assert len(files) == 3

    files = s(pattern=r'__init__\.py(?!c)', levels=None, exclude_folders=r'v?env')
    assert len(files) == 7

    files = s(pattern=r'__init__\.py(?!c)', levels=None, exclude_folders=['readers', 'v?env'])
    assert len(files) == 5

    files = s(pattern=r'__init__\.py(?!c)', levels=None, exclude_folders=['readers', 'writers', 'v?env'])
    assert len(files) == 4

    files = s(pattern=r'authors')
    assert len(files) == 0

    files = s(pattern=r'authors', regex_flags=re.IGNORECASE)
    assert len(files) == 1

    files = s(pattern=r'setup')
    assert len(files) == 2

    files = s(pattern=r'README', levels=None, exclude_folders='v?env')
    assert len(files) == 1

    files = s(pattern=r'README', levels=None, ignore_hidden_folders=False,
              exclude_folders=['.eggs', '.pytest_cache', '.cache', 'v?env'])
    assert len(files) == 1

    files = s(pattern=r'(^in|^auth)', levels=1, exclude_folders='htmlcov')
    assert len(files) == 3

    files = s(pattern=r'(^in|^auth)', levels=1, regex_flags=re.IGNORECASE, exclude_folders='htmlcov')
    assert len(files) == 4


def test_git_head():
    root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))

    head = utils.git_head(root_dir)
    assert len(head['hash']) == 40
    assert all(char.isalnum() for char in head['hash'])
    if sys.version_info.major > 2:
        assert isinstance(head['hash'], str)
    assert isinstance(head['datetime'], datetime.datetime)

    # can specify any directory within the version control hierarchy
    assert utils.git_head(os.path.join(root_dir, '.git')) == head
    assert utils.git_head(os.path.join(root_dir, 'msl', 'examples', 'io')) == head
    assert utils.git_head(os.path.dirname(__file__)) == head
    assert utils.git_head(os.path.join(root_dir, 'docs', '_api')) == head

    # a directory not under version control
    assert utils.git_head(tempfile.gettempdir()) is None


def test_checksum():
    path = os.path.join(os.path.dirname(__file__), 'samples', 'hdf5_sample.h5')
    sha256 = 'e5dad4f15335e603fd602c22bf9ddb71b3500f862905495d3d17e6159a463d2d'
    md5 = 'a46708df266595218db2ba06788c4695'

    # use a filename as a string
    assert isinstance(path, str)
    assert sha256 == utils.checksum(path, algorithm='sha256')
    assert md5 == utils.checksum(path, algorithm='md5')

    # use a filename as bytes
    path_as_bytes = path.encode()
    assert isinstance(path_as_bytes, bytes)
    assert sha256 == utils.checksum(path_as_bytes, algorithm='sha256')
    assert md5 == utils.checksum(path_as_bytes, algorithm='md5')

    # use a file pointer (binary mode)
    with open(path, mode='rb') as fp:
        assert sha256 == utils.checksum(fp, algorithm='sha256')
    with open(path, mode='rb') as fp:
        assert md5 == utils.checksum(fp, algorithm='md5')

    # use a byte buffer
    with open(path, mode='rb') as fp:
        buffer = BytesIO(fp.read())
    assert buffer.tell() == 0
    assert sha256 == utils.checksum(buffer, algorithm='sha256')
    assert buffer.tell() == 0
    assert md5 == utils.checksum(buffer, algorithm='md5')
    assert buffer.tell() == 0

    # ensure that all available algorithms can be used
    for algorithm in hashlib.algorithms_available:
        value = utils.checksum(path, algorithm=algorithm)
        assert isinstance(value, str)

    # file does not exist
    with pytest.raises((IOError, OSError), match='does_not_exist.txt'):
        utils.checksum('/the/file/does_not_exist.txt')
    with pytest.raises((IOError, OSError), match='does_not_exist.txt'):
        utils.checksum(b'/the/file/does_not_exist.txt')

    # invalid type
    with pytest.raises(TypeError):
        utils.checksum(None)
    with pytest.raises(TypeError):
        utils.checksum(bytearray(buffer.getvalue()))
    with pytest.raises(TypeError):
        utils.checksum(memoryview(buffer.getvalue()))

    # unsupported algorithm
    with pytest.raises(ValueError, match=r'unsupported'):
        utils.checksum(path, algorithm='invalid')


def test_get_basename():
    paths = [
        '/a/b/c/d/e/file.dat',
        'file.dat',
        '/a/file.dat',
        u'/something/file.dat',
        'file://a.b.c.d/folder/file.dat',
    ]
    if sys.platform == 'win32':
        paths.extend([
            'C:\\a\\b\\c\\d\\e\\file.dat',
            r'C:\a\file.dat',
            u'D:/file.dat',
            r'\\a.b.c.d\folder\file.dat',
            '\\\\a.b.c.d\\folder\\file.dat',
        ])
    for path in paths:
        assert utils.get_basename(path) == 'file.dat'
        assert utils.get_basename(path.encode()) == b'file.dat'

    assert utils.get_basename(StringIO(u'hello')) == u'StringIO'
    assert utils.get_basename(BytesIO(b'hello')) == 'BytesIO'

    with open(__file__, mode='rt') as fp:
        assert utils.get_basename(fp) == 'test_utils.py'


def test_copy():

    def check_stat(dest):
        src_stat = os.stat(__file__)
        dst_stat = os.stat(dest)
        for attr in dir(src_stat):
            if not attr.startswith('st_'):
                continue
            if attr in ['st_ino', 'st_ctime', 'st_ctime_ns', 'st_birthtime']:  # these will never be equal
                continue
            src_value = getattr(src_stat, attr)
            dst_value = getattr(dst_stat, attr)
            if attr == 'st_file_attributes':
                # on Windows the FILE_ATTRIBUTE_NOT_CONTENT_INDEXED attribute may not be copied
                if (src_value != dst_value) and (src_value + 0x2000 != dst_value):
                    return False
            elif 'time' in attr:  # times can be approximate
                if attr.endswith('ns'):
                    if abs(src_value - dst_value) > 1e4:
                        return False
                elif abs(src_value - dst_value) > 1e-4:
                    return False
            elif attr == 'st_dev' and sys.platform == 'win32' and os.getenv('GITHUB_ACTIONS'):
                # the ST_DEV values are not equal if the tests are run
                # via GitHub Actions and the OS is Windows
                pass
            elif src_value != dst_value:
                return False
        return True

    # make sure there is no remnant file from a previously-failed test
    if os.path.isfile(os.path.join(tempfile.gettempdir(), 'test_utils.py')):
        os.remove(os.path.join(tempfile.gettempdir(), 'test_utils.py'))

    # source file does not exist
    for item in [r'/the/file/does_not_exist.txt', r'/the/file/does_not_exist', r'does_not_exist']:
        with pytest.raises((IOError, OSError), match=item):
            utils.copy(item, '')

    # destination invalid
    with pytest.raises((IOError, OSError), match=r"''"):
        utils.copy(__file__, '')

    # copy (with metadata) to a directory that already exists
    dst = utils.copy(__file__, tempfile.gettempdir())
    assert dst == os.path.join(tempfile.gettempdir(), 'test_utils.py')
    assert check_stat(dst)
    assert utils.checksum(__file__) == utils.checksum(dst)

    # destination already exists
    with pytest.raises(OSError, match=r'Will not overwrite'):
        utils.copy(__file__, dst)
    with pytest.raises(OSError, match=r'Will not overwrite'):
        utils.copy(__file__, tempfile.gettempdir())

    # can overwrite (with metadata), specify full path
    dst2 = utils.copy(__file__, dst, overwrite=True)
    assert dst2 == dst
    assert check_stat(dst2)
    assert utils.checksum(__file__) == utils.checksum(dst2)

    # can overwrite (without metadata), specify full path
    dst3 = utils.copy(__file__, dst, overwrite=True, include_metadata=False)
    assert dst3 == dst
    assert not check_stat(dst3)
    assert utils.checksum(__file__) == utils.checksum(dst3)

    # can overwrite (with metadata), specify directory only
    dst4 = utils.copy(__file__, tempfile.gettempdir(), overwrite=True)
    assert dst4 == dst
    assert check_stat(dst4)
    assert utils.checksum(__file__) == utils.checksum(dst4)

    os.remove(dst)

    # copy without metadata
    dst = utils.copy(__file__, tempfile.gettempdir(), include_metadata=False)
    assert dst == os.path.join(tempfile.gettempdir(), 'test_utils.py')
    assert not check_stat(dst)
    assert utils.checksum(__file__) == utils.checksum(dst)
    os.remove(dst)

    # copy (without metadata) but use a different destination basename
    destination = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()) + '.tmp')
    dst = utils.copy(__file__, destination, include_metadata=False)
    assert dst == destination
    assert not check_stat(dst)
    assert utils.checksum(__file__) == utils.checksum(dst)
    os.remove(dst)

    # copy to a directory that does not exist
    new_dirs = str(uuid.uuid4()).split('-')
    assert not os.path.isdir(os.path.join(tempfile.gettempdir(), new_dirs[0]))
    destination = os.path.join(tempfile.gettempdir(), *new_dirs)
    destination = os.path.join(destination, 'new_file.tmp')
    dst = utils.copy(__file__, destination)
    assert dst == destination
    assert check_stat(dst)
    assert utils.checksum(__file__) == utils.checksum(dst)
    shutil.rmtree(os.path.join(tempfile.gettempdir(), new_dirs[0]))


def test_remove_write_permissions():

    # create a new file
    path = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()) + '.tmp')
    with open(path, mode='wb') as fp:
        fp.write(b'hello')

    # set to rwxrwxrwx
    os.chmod(path, 0o777)

    mode = stat.S_IMODE(os.lstat(path).st_mode)
    if sys.platform == 'win32':
        assert mode == 0o666  # Windows does not have the Execute permission
    else:
        assert mode == 0o777

    # can still modify it
    with open(path, mode='ab') as fp:
        fp.write(b' world')
    with open(path, mode='rb') as fp:
        assert fp.read() == b'hello world'

    utils.remove_write_permissions(path)

    # the Read and Execute permissions are preserved
    mode = stat.S_IMODE(os.lstat(path).st_mode)
    if sys.platform == 'win32':
        assert mode == 0o444  # Windows does not have the Execute permission
    else:
        assert mode == 0o555

    # cannot open the file to modify it
    for m in ['wb', 'ab', 'wt', 'at', 'w+', 'w+b']:
        with pytest.raises((IOError, OSError)):
            open(path, mode=m)

    # cannot delete the file (only valid on Windows)
    if sys.platform == 'win32':
        with pytest.raises(OSError):
            os.remove(path)

    # can still read it
    with open(path, mode='rb') as fp:
        assert fp.read() == b'hello world'

    # set to rw--wxrw-
    os.chmod(path, 0o636)
    # remove and check permissions
    utils.remove_write_permissions(path)
    mode = stat.S_IMODE(os.lstat(path).st_mode)
    if sys.platform == 'win32':
        # Windows does not have the Execute permission
        # and if any of the Read permissions are enabled then it
        # is enabled for the User, Group and Others
        assert mode == 0o444
    else:
        assert mode == 0o414

    # clean up by deleting the file
    os.chmod(path, 0o777)
    os.remove(path)


def test_is_file_readable():
    assert utils.is_file_readable(__file__)

    for item in [None, '', __file__+'py', dict()]:
        assert not utils.is_file_readable(item)

    with pytest.raises(TypeError):
        utils.is_file_readable(None, strict=True)

    with pytest.raises((IOError, OSError)):
        utils.is_file_readable('', strict=True)

    with pytest.raises((IOError, OSError)):
        utils.is_file_readable(__file__+'py', strict=True)


def test_is_dir_accessible():
    assert utils.is_dir_accessible(os.path.dirname(__file__))

    for item in [None, '', __file__, dict()]:
        assert not utils.is_dir_accessible(item)

    with pytest.raises(TypeError):
        utils.is_dir_accessible(None, strict=True)

    with pytest.raises(OSError):
        utils.is_dir_accessible('', strict=True)

    with pytest.raises(OSError):
        utils.is_dir_accessible(__file__, strict=True)


def test_is_admin():
    assert isinstance(utils.is_admin(), bool)


@pytest.mark.skipif(utils.is_admin(), reason='don\'t run if already an admin')
@pytest.mark.skipif(os.name != 'nt', reason='non-Windows OS')
def test_run_as_admin():
    # Using verb=None as a keyword argument allows for testing the
    # 'run_as_admin' function without getting the UAC prompt,
    # but the test command must not require admin privileges.
    # Only test on a non-admin Windows session because the other cases:
    #   1) already an admin
    #   2) running on POSIX
    # are straightforward implementations and we don't want to test
    # the subprocess.check_output internals.

    with pytest.raises(ValueError, match=r'args and/or an executable$'):
        utils.run_as_admin()

    # Don't want Windows to return b'ECHO is on.\r\n'
    out = utils.run_as_admin(args='echo 1', verb=None)
    assert out == b'1 \r\n'
    out = utils.run_as_admin(args=['echo', '1'], verb=None)
    assert out == b'1 \r\n'

    # no extra space after "hi"
    out = utils.run_as_admin(args='echo hi', verb=None)
    assert out == b'hi\r\n'

    # pass in the keyword argument "args=" as a positional argument
    out = utils.run_as_admin('echo hi', verb=None)
    assert out == b'hi\r\n'

    # run a batch file with spaces in the file path and in the arguments
    file = os.path.join(tempfile.gettempdir(), 'msl io batch script.bat')
    with open(file, mode='wt') as fp:
        fp.write('@ECHO OFF\r\n')
        for i in range(1, 6):
            fp.write('echo %{}\r\n'.format(i))

    expected = b'ECHO is off.\r\nECHO is off.\r\nECHO is off.\r\nECHO is off.\r\nECHO is off.\r\n'
    out = utils.run_as_admin(args='"{}"'.format(file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file], verb=None)
    assert out == expected
    out = utils.run_as_admin(executable=file, verb=None)
    assert out == expected

    expected = b'pi\r\n-p\r\nECHO is off.\r\nECHO is off.\r\nECHO is off.\r\n'
    out = utils.run_as_admin(args='"{}" pi -p'.format(file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file, 'pi', '-p'], verb=None)
    assert out == expected
    out = utils.run_as_admin(args=['pi', '-p'], executable=file, verb=None)
    assert out == expected

    expected = b'pi\r\n-p\r\n"hel lo"\r\nECHO is off.\r\nECHO is off.\r\n'
    out = utils.run_as_admin(args='"{}" pi -p "hel lo"'.format(file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file, 'pi', '-p', 'hel lo'], verb=None)
    assert out == expected
    out = utils.run_as_admin(args=['pi', '-p', 'hel lo'], executable=file, verb=None)
    assert out == expected

    expected = b'pi\r\n-p\r\n"hel lo"\r\n6\r\n"last parameter received"\r\n'
    out = utils.run_as_admin(args='"{}" pi -p "hel lo" 6 "last parameter received"'.format(file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file, 'pi', '-p', 'hel lo', '6', 'last parameter received'], verb=None)
    assert out == expected
    out = utils.run_as_admin(args=['pi', '-p', 'hel lo', '6', 'last parameter received'], executable=file, verb=None)
    assert out == expected

    # change directory
    out = utils.run_as_admin(
        args=[os.path.basename(file), 'pi', '-p', 'hel lo', '6', 'last parameter received'],
        cwd=os.path.dirname(file),
        verb=None
    )
    assert out == expected

    os.remove(file)

    # call the Python interpreter
    expected = subprocess.check_output([sys.executable, '-VV'])
    out = utils.run_as_admin(args='-VV', executable=sys.executable, verb=None)
    assert out == expected
    out = utils.run_as_admin(args='{} -VV'.format(sys.executable), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[sys.executable, '-VV'], verb=None)
    assert out == expected
    out = utils.run_as_admin(args=['-VV'], executable=sys.executable, verb=None)
    assert out == expected

    # run a python script
    file = os.path.join(tempfile.gettempdir(), 'msl_io_admin_test.py')
    with open(file, mode='wt') as fp:
        fp.write('from __future__ import print_function\r\n')
        fp.write('import sys\r\n')
        # additional packages must be available since msl-io depends on them
        fp.write('import numpy\r\n')
        fp.write('print(sys.executable)\r\n')
        fp.write('print(sys.argv[1:])\r\n')
        fp.write('print("written to stderr", file=sys.stderr)\r\n')

    # no arguments
    expected = '{}\r\n[]\r\n'.format(sys.executable).encode()
    out = utils.run_as_admin(args=file, executable=sys.executable, verb=None)
    assert out == expected
    out = utils.run_as_admin(args='{} {}'.format(sys.executable, file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file], executable=sys.executable, verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[sys.executable, file], verb=None)
    assert out == expected

    # with arguments
    expected = "{}\r\n['1', 'x=5', 'a b c d', 'e']\r\n".format(sys.executable).encode()
    out = utils.run_as_admin(args='{} 1 x=5 "a b c d" e'.format(file), executable=sys.executable, verb=None)
    assert out == expected
    out = utils.run_as_admin(args='{} {} 1 x=5 "a b c d" e'.format(sys.executable, file), verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[file, '1', 'x=5', 'a b c d', 'e'], executable=sys.executable, verb=None)
    assert out == expected
    out = utils.run_as_admin(args=[sys.executable, file, '1', 'x=5', 'a b c d', 'e'], verb=None)
    assert out == expected

    # change directory
    expected = "{}\r\n['1']\r\n".format(sys.executable).encode()
    out = utils.run_as_admin(args=[sys.executable, os.path.basename(file), '1'], cwd=os.path.dirname(file), verb=None)
    assert out == expected

    # capture_stderr=True
    out = utils.run_as_admin(args=[sys.executable, file], capture_stderr=True, verb=None)
    assert b'written to stderr' in out

    os.remove(file)

    # raise some exceptions
    with pytest.raises(PermissionError):
        utils.run_as_admin('sc create MSL-IO-TEST binPath= "C:\\hello world\\dummy.exe"', verb=None)

    with pytest.raises(PermissionError):
        utils.run_as_admin(['sc', 'create', 'MSL-IO-TEST', 'binPath=', 'C:\\hello world\\dummy.exe'], verb=None)

    with pytest.raises(OSError, match=r'Set capture_stderr=True to see if more information is available.$'):
        utils.run_as_admin('doesnotexist.exe', verb=None)

    with pytest.raises(OSError, match=r"'doesnotexist.exe' is not recognized as an internal or external command"):
        utils.run_as_admin('doesnotexist.exe', capture_stderr=True, verb=None)

    with pytest.raises(OSError, match=r'Set show=False to capture the stdout stream.$'):
        utils.run_as_admin([sys.executable, '-c', '1/0'], show=True, verb=None)

    with pytest.raises(OSError, match=r'Set capture_stderr=True to see if more information is available.$'):
        utils.run_as_admin([sys.executable, '-c', '1/0'], verb=None)

    with pytest.raises(OSError, match=r'Set capture_stderr=True to see if more information is available.\nHello!$'):
        utils.run_as_admin([sys.executable, '-c', 'print("Hello!");1/0'], verb=None)

    with pytest.raises(OSError, match=r'ZeroDivisionError:'):
        utils.run_as_admin([sys.executable, '-c', '1/0'], capture_stderr=True, verb=None)


def test_prepare_email():
    temp = os.path.join(tempfile.gettempdir(), '793a7e5d-7e0e-4049-9e9b-f4383b7bb96a.tmp')

    def create(lines):
        with open(temp, mode='wt') as fp:
            fp.write('\n'.join(lines))

    with pytest.raises((OSError, IOError)):
        utils._prepare_email('does-not-exist.ini', '', '')

    create(['[smtp]', '[gmail]'])
    with pytest.raises(ValueError, match='Cannot specify both'):
        utils._prepare_email(temp, '', None)

    create(['[unknown]'])
    with pytest.raises(ValueError, match='Must create either'):
        utils._prepare_email(temp, '', None)

    for item in (['[smtp]'], ['[smtp]', 'host=hostname'], ['[smtp]', 'port=25']):
        create(item)
        with pytest.raises(ValueError, match="Must specify the 'host' and 'port'"):
            utils._prepare_email(temp, '', None)

    create(['[smtp]', 'port=not-an-int'])
    with pytest.raises(ValueError, match='invalid literal for int()'):
        utils._prepare_email(temp, '', None)

    create(['[smtp]', 'host=smtp.example.com', 'port=25'])
    cfg = utils._prepare_email(temp, '', None)
    assert cfg == {'type': 'smtp', 'to': [''], 'from': '', 'host': 'smtp.example.com',
                   'port': 25, 'starttls': None, 'username': None, 'password': None}

    create(['[smtp]', 'host=h', 'port=1'])
    cfg = utils._prepare_email(temp, 'name', None)
    assert cfg == {'type': 'smtp', 'to': ['name'], 'from': 'name', 'host': 'h',
                   'port': 1, 'starttls': None, 'username': None, 'password': None}

    create(['[smtp]', 'host=h', 'port=1', 'domain=domain'])
    cfg = utils._prepare_email(temp, 'name', None)
    assert cfg == {'type': 'smtp', 'to': ['name@domain'], 'from': 'name@domain', 'host': 'h',
                   'port': 1, 'starttls': None, 'username': None, 'password': None}

    create(['[smtp]', 'host=h', 'port=1', 'domain=@domain'])
    cfg = utils._prepare_email(temp, ['name0', 'name1', 'name2'], None)
    assert cfg == {'type': 'smtp', 'to': ['name0@domain', 'name1@domain', 'name2@domain'],
                   'from': 'name0@domain', 'host': 'h', 'port': 1, 'starttls': None,
                   'username': None, 'password': None}

    create(['[smtp]', 'host=h', 'port=1', 'domain=@domain'])
    cfg = utils._prepare_email(temp, 'name@mail.com', None)
    assert cfg == {'type': 'smtp', 'to': ['name@mail.com'], 'from': 'name@mail.com',
                   'host': 'h', 'port': 1, 'starttls': None, 'username': None,
                   'password': None}

    create(['[smtp]', 'host=h', 'port=1', 'domain=@domain'])
    cfg = utils._prepare_email(temp, ['you@mail.com', 'me'], 'me')
    assert cfg == {'type': 'smtp', 'to': ['you@mail.com', 'me@domain'],
                   'from': 'me@domain', 'host': 'h', 'port': 1, 'starttls': None,
                   'username': None, 'password': None}

    for item in ('yes', 'YES', '1', 'true', 'True', 'on', 'On'):
        create(['[smtp]', 'host=h', 'port=1', 'starttls={}'.format(item)])
        cfg = utils._prepare_email(temp, 'you', 'me')
        assert cfg == {'type': 'smtp', 'to': ['you'], 'from': 'me', 'host': 'h',
                       'port': 1, 'starttls': True, 'username': None, 'password': None}

    for item in ('no', 'No', '0', 'false', 'False', 'off', 'Off'):
        create(['[smtp]', 'host=h', 'port=1', 'starttls={}'.format(item)])
        cfg = utils._prepare_email(temp, 'you', 'me')
        assert cfg == {'type': 'smtp', 'to': ['you'], 'from': 'me', 'host': 'h',
                       'port': 1, 'starttls': False, 'username': None, 'password': None}

    create(['[smtp]', 'host=h', 'port=1', 'username=user'])
    with pytest.raises(ValueError, match="Must specify the 'password'"):
        utils._prepare_email(temp, '', None)

    create(['[smtp]', 'host=h', 'port=1', 'password=pw'])
    with pytest.raises(ValueError, match="Must specify the 'username'"):
        utils._prepare_email(temp, '', None)

    create(['[smtp]', 'host=h', 'port=1', 'starttls=0', 'username=uname', 'password=pw'])
    cfg = utils._prepare_email(temp, 'you', 'me@mail')
    assert cfg == {'type': 'smtp', 'to': ['you'], 'from': 'me@mail', 'host': 'h',
                   'port': 1, 'starttls': False, 'username': 'uname', 'password': 'pw'}

    create(['[gmail]'])
    cfg = utils._prepare_email(temp, '', None)
    assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': None}
    cfg = utils._prepare_email(temp, 'me', None)
    assert cfg == {'type': 'gmail', 'to': ['me'], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': None}

    create(['[gmail]', 'domain=domain'])
    cfg = utils._prepare_email(temp, 'me', None)
    assert cfg == {'type': 'gmail', 'to': ['me'], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': None}

    create(['[gmail]', 'account=mine', 'domain=@gmail.com'])
    cfg = utils._prepare_email(temp, 'you', 'me')
    assert cfg == {'type': 'gmail', 'to': ['you@gmail.com'], 'from': 'me',
                   'account': 'mine', 'credentials': None, 'scopes': None}

    create(['[gmail]', 'credentials=path/to/oauth'])
    cfg = utils._prepare_email(temp, ['email@corp.com', 'me'], None)
    assert cfg == {'type': 'gmail', 'to': ['email@corp.com', 'me'], 'from': 'me',
                   'account': None, 'credentials': 'path/to/oauth', 'scopes': None}

    create(['[gmail]', 'credentials=path\\to\\oauth', 'account=work', 'domain=ignored'])
    cfg = utils._prepare_email(temp, 'name@gmail.com', 'name@email.com')
    assert cfg == {'type': 'gmail', 'to': ['name@gmail.com'], 'from': 'name@email.com',
                   'account': 'work', 'credentials': 'path\\to\\oauth', 'scopes': None}

    create(['[gmail]', 'credentials=path\\to\\oauth', 'account=work', 'domain=domain'])
    cfg = utils._prepare_email(temp, ['a', 'b', 'c@corp.net'], None)
    assert cfg == {'type': 'gmail', 'to': ['a@domain', 'b@domain', 'c@corp.net'],
                   'from': 'me', 'account': 'work', 'credentials': 'path\\to\\oauth',
                   'scopes': None}

    create(['[gmail]', 'scopes=a'])
    cfg = utils._prepare_email(temp, '', None)
    assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': ['a']}

    create(['[gmail]', 'scopes = ', ' gmail', ' gmail.send', ' g',
            ' gmail.metadata', '', '', '', 'account = work'])
    cfg = utils._prepare_email(temp, '', '')
    assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                   'account': 'work', 'credentials': None,
                   'scopes': ['gmail', 'gmail.send', 'g', 'gmail.metadata']}

    # file-like objects
    with open(temp, mode='rt') as fp:
        cfg = utils._prepare_email(fp, '', '')
        assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                       'account': 'work', 'credentials': None,
                       'scopes': ['gmail', 'gmail.send', 'g', 'gmail.metadata']}

    cfg = utils._prepare_email(StringIO(u'[gmail]'), '', None)
    assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': None}

    cfg = utils._prepare_email(BytesIO(b'[gmail]'), '', None)
    assert cfg == {'type': 'gmail', 'to': [''], 'from': 'me',
                   'account': None, 'credentials': None, 'scopes': None}

    os.remove(temp)



import io
import os
import sys
import tempfile
import uuid
from datetime import datetime

import pytest

try:
    from googleapiclient.errors import HttpError
except ImportError:
    HttpError = Exception

from msl.io.constants import IS_PYTHON2
from msl.io.google_api import GCell
from msl.io.google_api import GCellType
from msl.io.google_api import GDrive
from msl.io.google_api import GMail
from msl.io.google_api import GSheets
from msl.io.google_api import GValueOption

# all Google API tests require the necessary "token.json" file to be
# available for a specific Google user's account
try:
    # dr -> drive, readonly
    dr = GDrive(account='testing', read_only=True)
except:
    dr = None

try:
    # dw -> drive, writable
    dw = GDrive(account='testing', read_only=False)
except:
    dw = None

try:
    # sr -> sheets, readonly
    sr = GSheets(account='testing', read_only=True)
except:
    sr = None

try:
    # sw -> sheets, writeable
    sw = GSheets(account='testing', read_only=False)
except:
    sw = None

try:
    gmail = GMail(account='testing')
except:
    gmail = None

skipif_no_gdrive_readonly = pytest.mark.skipif(
    dr is None, reason='No GDrive readonly token'
)

skipif_no_gdrive_writeable = pytest.mark.skipif(
    dw is None, reason='No GDrive writable token'
)

skipif_no_sheets_readonly = pytest.mark.skipif(
    sr is None, reason='No GSheets readonly token'
)

skipif_no_sheets_writeable = pytest.mark.skipif(
    sw is None, reason='No GSheets writeable token'
)

skipif_no_gmail = pytest.mark.skipif(
    gmail is None, reason='No Gmail token'
)

IS_WINDOWS = sys.platform == 'win32'


@skipif_no_sheets_readonly
def test_gsheets_sheet_names():
    # MSL/msl-io-testing/empty-5.gsheet
    names = sr.sheet_names('1Ua15pRGUH5qoU0c3Ipqrkzi9HBlm3nzqCn5O1IONfCY')
    assert len(names) == 5
    assert 'Sheet1' in names
    assert 'Sheet2' in names
    assert 'Sheet3' in names
    assert 'Sheet4' in names
    assert 'Sheet5' in names

    # MSL/msl-io-testing/f 1/f2/sub folder 3/lab environment.gsheet
    names = sr.sheet_names('1FwzsFgN7w-HZXOlUAEMVMSOGpNHCj5NXvH6Xl7LyLp4')
    assert len(names) == 1
    assert 'Sensor_1' in names

    # table.gsheet
    names = sr.sheet_names('1Q0TAgnw6AJQWkLMf8V3qEhEXuCEXTFAc95cEcshOXnQ')
    assert len(names) == 6
    assert 'StartA1' in names
    assert 'StartH22' in names
    assert 'header only' in names
    assert 'empty' in names
    assert 'column' in names
    assert 'row' in names


@skipif_no_sheets_readonly
def test_gsheets_values():
    # MSL/msl-io-testing/empty-5
    empty_id = '1Ua15pRGUH5qoU0c3Ipqrkzi9HBlm3nzqCn5O1IONfCY'

    # MSL/msl-io-testing/f 1/f2/sub folder 3/lab environment
    lab_id = '1FwzsFgN7w-HZXOlUAEMVMSOGpNHCj5NXvH6Xl7LyLp4'

    # more than 1 sheet exists
    with pytest.raises(ValueError, match=r'You must specify a sheet name:'):
        sr.values(empty_id)

    # empty sheets are okay
    for name in sr.sheet_names(empty_id):
        values = sr.values(empty_id, sheet=name)
        assert isinstance(values, list)
        assert not values

        # specifying the cells in an empty sheet is okay
        values = sr.values(empty_id, sheet=name, cells='A2:Z10')
        assert isinstance(values, list)
        assert not values

    # only 1 sheet exists, therefore we do not need to specify
    # a value for the 'sheet' kwarg since it is determined automatically
    expected = [
        ['Timestamp', 'Temperature', 'Humidity'],
        ['2021-04-03 12:36:10', '20.33', '49.82'],
        ['2021-04-03 12:37:10', '20.23', '46.06'],
        ['2021-04-03 12:38:10', '20.41', '47.06'],
        ['2021-04-03 12:39:10', '20.29', '48.32']
    ]
    values = sr.values(lab_id)
    assert values == expected

    values = sr.values(lab_id, row_major=False)
    assert values == [
        ['Timestamp', '2021-04-03 12:36:10', '2021-04-03 12:37:10', '2021-04-03 12:38:10', '2021-04-03 12:39:10'],
        ['Temperature', '20.33', '20.23', '20.41', '20.29'],
        ['Humidity', '49.82', '46.06', '47.06', '48.32']
    ]

    values = sr.values(lab_id, cells='B2:C4', value_option='FORMATTED_VALUE')
    assert values == [['20.33', '49.82'], ['20.23', '46.06'], ['20.41', '47.06']]

    values = sr.values(lab_id, cells='B:B', value_option='UNFORMATTED_VALUE')
    assert values == [['Temperature'], [20.33], [20.23], [20.41], [20.29]]

    values = sr.values(lab_id, cells='B:C', value_option=GValueOption.UNFORMATTED)
    assert values == [['Temperature', 'Humidity'], [20.33, 49.82], [20.23, 46.06], [20.41, 47.06], [20.29, 48.32]]

    values = sr.values(lab_id, cells='A2:C2')
    assert values == [expected[1]]


@skipif_no_sheets_readonly
def test_gsheets_to_datetime():
    expected = [
        ['Timestamp', datetime(2021, 4, 3, 12, 36, 10), datetime(2021, 4, 3, 12, 37, 10),
         datetime(2021, 4, 3, 12, 38, 10), datetime(2021, 4, 3, 12, 39, 10)],
        ['Temperature', 20.33, 20.23, 20.41, 20.29],
        ['Humidity', 49.82, 46.06, 47.06, 48.32]
    ]

    # MSL/msl-io-testing/f 1/f2/sub folder 3/lab environment
    lab_id = '1FwzsFgN7w-HZXOlUAEMVMSOGpNHCj5NXvH6Xl7LyLp4'
    values = sr.values(lab_id, value_option='UNFORMATTED_VALUE', row_major=False)
    values[0][1:] = [sr.to_datetime(t) for t in values[0][1:]]
    assert values == expected

    values = sr.values(lab_id, value_option='UNFORMATTED_VALUE',
                       datetime_option='FORMATTED_STRING', row_major=False)
    expected[0][1:] = [str(t) for t in expected[0][1:]]
    assert values == expected


@skipif_no_sheets_readonly
def test_gsheets_cells():
    # MSL/msl-io-testing/empty-5
    empty_id = '1Ua15pRGUH5qoU0c3Ipqrkzi9HBlm3nzqCn5O1IONfCY'

    # data-types
    datatypes_id = '1zMO4wk0IPC9I57dR5WoPTzlOX6g5-AcnwGFOEHrhIHU'

    # invalid spreadsheet_id
    with pytest.raises(HttpError):
        sr.cells(empty_id[:-1] + 'A')

    # valid spreadsheet_id, invalid sheet name
    with pytest.raises(HttpError):
        sr.cells(datatypes_id, ranges='invalid')
    with pytest.raises(HttpError):
        sr.cells(datatypes_id, ranges=['invalid'])

    assert sr.cells(empty_id) == {'Sheet1': [], 'Sheet2': [], 'Sheet3': [], 'Sheet4': [], 'Sheet5': []}

    assert sr.cells(empty_id, ranges='Sheet1') == {'Sheet1': []}
    assert sr.cells(empty_id, ranges=['Sheet1']) == {'Sheet1': []}
    assert sr.cells(empty_id, ranges=['Sheet1', 'Sheet5']) == {'Sheet1': [], 'Sheet5': []}
    assert sr.cells(empty_id, ranges=['Sheet1', 'Sheet3!B7:ZZ99']) == {'Sheet1': [], 'Sheet3': []}

    cells = sr.cells(datatypes_id)
    values = cells['Data Types']
    assert len(values) == 18

    assert len(values[0]) == 6
    assert values[0][0] == GCell(value='Automatic', type=GCellType.STRING, formatted='Automatic')
    assert values[0][1] == GCell(value=1.23, type=GCellType.NUMBER, formatted='1.23')
    assert values[0][2] == GCell(value='string', type=GCellType.STRING, formatted='string')
    assert values[0][3] == GCell(value=1, type=GCellType.NUMBER, formatted='1')
    assert values[0][4] == GCell(value='0123456789', type=GCellType.STRING, formatted='0123456789')
    assert values[0][5] == GCell(value=36982, type=GCellType.DATE, formatted='1 April 2001')

    assert len(values[1]) == 3
    assert values[1][0] == GCell(value='Plain text', type=GCellType.STRING, formatted='Plain text')
    assert values[1][1] == GCell(value='a b c d', type=GCellType.STRING, formatted='a b c d')
    assert values[1][2] == GCell(value='34', type=GCellType.STRING, formatted='34')

    assert len(values[2]) == 2
    assert values[2][0] == GCell(value='Number', type=GCellType.STRING, formatted='Number')
    assert values[2][1] == GCell(value=1234.56789, type=GCellType.NUMBER, formatted='1,234.57')

    assert len(values[3]) == 2
    assert values[3][0] == GCell(value='Percent', type=GCellType.STRING, formatted='Percent')
    assert values[3][1] == GCell(value=0.542, type=GCellType.PERCENT, formatted='54.20%')

    assert len(values[4]) == 2
    assert values[4][0] == GCell(value='Scientific', type=GCellType.STRING, formatted='Scientific')
    assert values[4][1] == GCell(value=0.00321, type=GCellType.SCIENTIFIC, formatted='3.21E-03')

    assert len(values[5]) == 3
    assert values[5][0] == GCell(value='Accounting', type=GCellType.STRING, formatted='Accounting')
    assert values[5][1] == GCell(value=99.95, type=GCellType.NUMBER, formatted=' $ 99.95 ')
    assert values[5][2] == GCell(value=-23.45, type=GCellType.NUMBER, formatted=' $ (23.45)')

    assert len(values[6]) == 3
    assert values[6][0] == GCell(value='Financial', type=GCellType.STRING, formatted='Financial')
    assert values[6][1] == GCell(value=1.23, type=GCellType.NUMBER, formatted='1.23')
    assert values[6][2] == GCell(value=-1.23, type=GCellType.NUMBER, formatted='(1.23)')

    assert len(values[7]) == 3
    assert values[7][0] == GCell(value='Currency', type=GCellType.STRING, formatted='Currency')
    assert values[7][1] == GCell(value=99.95, type=GCellType.CURRENCY, formatted='$99.95')
    assert values[7][2] == GCell(value=-1.99, type=GCellType.CURRENCY, formatted='-$1.99')

    assert len(values[8]) == 3
    assert values[8][0] == GCell(
        value='Currency (rounded)', type=GCellType.STRING, formatted='Currency (rounded)')
    assert values[8][1] == GCell(value=99.95, type=GCellType.CURRENCY, formatted='$100')
    assert values[8][2] == GCell(value=-1.99, type=GCellType.CURRENCY, formatted='-$2')

    assert len(values[9]) == 2
    assert values[9][0] == GCell(value='Date', type=GCellType.STRING, formatted='Date')
    assert values[9][1] == GCell(value=17738, type=GCellType.DATE, formatted='24/07/1948')

    assert len(values[10]) == 3
    assert values[10][0] == GCell(value='Time', type=GCellType.STRING, formatted='Time')
    assert values[10][1] == GCell(value=0.2661689814814815, type=GCellType.TIME, formatted='06:23:17')
    assert values[10][2] == GCell(value=0.7378356481481482, type=GCellType.TIME, formatted='17:42:29')

    assert len(values[11]) == 2
    assert values[11][0] == GCell(value='Date time', type=GCellType.STRING, formatted='Date time')
    assert values[11][1] == GCell(
        value=34736.4303472222222222, type=GCellType.DATE_TIME, formatted='06/02/1995 10:19:42')

    assert len(values[12]) == 2
    assert values[12][0] == GCell(value='Duration', type=GCellType.STRING, formatted='Duration')
    assert values[12][1] == GCell(value=1.000023148148148, type=GCellType.TIME, formatted='24:00:02')

    assert len(values[13]) == 2
    assert values[13][0] == GCell(value='Formula', type=GCellType.STRING, formatted='Formula')
    assert values[13][1] == GCell(value=6.747908247937978, type=GCellType.SCIENTIFIC, formatted='6.75E+00')

    assert len(values[14]) == 3
    assert values[14][0] == GCell(value='Error', type=GCellType.STRING, formatted='Error')
    assert values[14][1] == GCell(
        value='#DIV/0! (Function DIVIDE parameter 2 cannot be zero.)',
        type=GCellType.ERROR,
        formatted='#DIV/0!')
    assert values[14][2] == GCell(
        value="#VALUE! (Function MULTIPLY parameter 2 expects number values. "
              "But 'Currency' is a text and cannot be coerced to a number.)",
        type=GCellType.ERROR,
        formatted='#VALUE!')

    assert len(values[15]) == 3
    assert values[15][0] == GCell(value='Empty', type=GCellType.STRING, formatted='Empty')
    assert values[15][1] == GCell(value=None, type=GCellType.EMPTY, formatted='')
    assert values[15][2] == GCell(
        value='<== keep B16 empty', type=GCellType.STRING, formatted='<== keep B16 empty')

    assert len(values[16]) == 3
    assert values[16][0] == GCell(value='Boolean', type=GCellType.STRING, formatted='Boolean')
    assert values[16][1] == GCell(value=True, type=GCellType.BOOLEAN, formatted='TRUE')
    assert values[16][2] == GCell(value=False, type=GCellType.BOOLEAN, formatted='FALSE')

    assert len(values[17]) == 2
    assert values[17][0] == GCell(value='Custom', type=GCellType.STRING, formatted='Custom')
    assert values[17][1] == GCell(value=12345.6789, type=GCellType.NUMBER, formatted='12345 55/81')


@skipif_no_sheets_writeable
@skipif_no_gdrive_writeable
def test_gsheets_create_move_delete():
    sid = sw.create('no-sheet-names')
    assert sw.sheet_names(sid) == ('Sheet1',)
    dw.delete(sid)

    sid = sw.create('three-sheet-names', sheet_names=['a', 'bb', 'ccc'])
    assert dw.path(sid) == 'My Drive/three-sheet-names'
    assert sw.sheet_names(sid) == ('a', 'bb', 'ccc')
    fid = dw.create_folder('eat/more/fruit')
    dw.move(sid, fid)
    assert dw.path(sid) == 'My Drive/eat/more/fruit/three-sheet-names'
    dw.delete(dw.folder_id('My Drive/eat'))


@skipif_no_gdrive_readonly
def test_gdrive_shared_drives():
    assert dr.shared_drives() == {}


@skipif_no_gdrive_readonly
def test_gdrive_folder_id_exception():
    # the folder does not exist
    folders = [
        'DoesNotExist',
        '/Google Drive/MSL/DoesNotExist',
    ]
    if IS_WINDOWS:
        folders.append(r'C:\Users\username\Google Drive\MSL\DoesNotExist')
    for folder in folders:
        with pytest.raises(OSError, match=r'Not a valid Google Drive folder'):
            dr.folder_id(folder)

    # specified a valid file (which is not a folder)
    files = [
        'Single-Photon Generation and Detection.pdf',
        'MSL/msl-io-testing/unique'
    ]
    for file in files:
        with pytest.raises(OSError, match=r'Not a valid Google Drive folder'):
            dr.folder_id(file)

    # specify an invalid parent ID
    assert dr.folder_id('MSL') == '14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi'
    with pytest.raises(HttpError):
        dr.folder_id('MSL', parent_id='INVALID_Kmkjo9aCQGysOsxwkTtpoJODi')


@skipif_no_gdrive_readonly
def test_gdrive_folder_id():
    # relative to the root folder
    assert dr.folder_id('') == 'root'
    assert dr.folder_id('/') == 'root'
    assert dr.folder_id('Google Drive') == 'root'
    if IS_WINDOWS:
        assert dr.folder_id('C:\\Users\\username\\Google Drive') == 'root'
        assert dr.folder_id(r'D:\Google Drive') == 'root'

    assert dr.folder_id('MSL') == '14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi'
    assert dr.folder_id('MSL/msl-io-testing') == '1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C'
    assert dr.folder_id('MSL/msl-io-testing/f 1/f2') == '1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN'
    assert dr.folder_id('/Google Drive/MSL') == '14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi'
    assert dr.folder_id('Google Drive/MSL/msl-io-testing/f 1') == '1mhQ_9iVF5AhbUb7Lq77qzuBbvZr150X9'
    if IS_WINDOWS:
        assert dr.folder_id('C:\\Users\\username\\Google Drive\\MSL') == '14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi'
        assert dr.folder_id(r'MSL\msl-io-testing\f 1\f2\sub folder 3') == '1wLAPHCOphcOITR37b8UB88eFW_FzeNQB'

    # relative to a parent folder
    assert dr.folder_id('msl-io-testing', parent_id='14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi') == '1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C'
    assert dr.folder_id('msl-io-testing/f 1/f2', parent_id='14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi') == '1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN'
    assert dr.folder_id('f 1/f2', parent_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C') == '1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN'
    assert dr.folder_id('f2', parent_id='1mhQ_9iVF5AhbUb7Lq77qzuBbvZr150X9') == '1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN'
    assert dr.folder_id('sub folder 3', parent_id='1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN') == '1wLAPHCOphcOITR37b8UB88eFW_FzeNQB'


@skipif_no_gdrive_readonly
def test_gdrive_file_id_exception():
    # file does not exist
    files = [
        'DoesNotExist',
        '/home/username/Google Drive/DoesNotExist.txt',
    ]
    if IS_WINDOWS:
        files.append(r'C:\Users\username\Google Drive\DoesNotExist.txt')
    for file in files:
        with pytest.raises(OSError, match=r'Not a valid Google Drive file'):
            dr.file_id(file)

    # specified a valid folder (which is not a file)
    folders = [
        'MSL',
        '/Google Drive/MSL',
    ]
    if IS_WINDOWS:
        folders.append(r'C:\Users\username\Google Drive\MSL')
    for folder in folders:
        with pytest.raises(OSError, match=r'Not a valid Google Drive file'):
            dr.file_id(folder)

    # specify an invalid parent ID
    assert dr.file_id('unique', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C') == '1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395'
    with pytest.raises(HttpError):
        dr.file_id('unique', folder_id='INVALID_NCuTWxmABs-w7JenftaLGAG9C')


@skipif_no_gdrive_readonly
def test_gdrive_file_id():
    # relative to the root folder
    files = {
        'Single-Photon Generation and Detection.pdf': '11yaxZH93B0IhQZwfCeo2dXb-Iduh-4dS',
        'MSL/msl-io-testing/unique': '1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395',
        '/Google Drive/Single-Photon Generation and Detection.pdf': '11yaxZH93B0IhQZwfCeo2dXb-Iduh-4dS',
    }
    if IS_WINDOWS:
        files[r'C:\Users\username\Google Drive\MSL\msl-io-testing\unique'] = '1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395'
        files['MSL\\msl-io-testing\\f 1\\f2\\New Text Document.txt'] = '1qW1QclelxZtJtKMigCgGH4ST3QoJ9zuP'

    for file, id_ in files.items():
        assert dr.file_id(file) == id_

    # relative to a parent folder
    assert dr.file_id('unique', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C') == '1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395'
    assert dr.file_id('msl-io-testing/unique', folder_id='14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi') == '1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395'
    assert dr.file_id('file.txt', folder_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB') == '1CDS3cWDItXB1uLCPGq0uy6OJAngkmNoD'
    assert dr.file_id('f 1/f2/New Text Document.txt', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C') == '1qW1QclelxZtJtKMigCgGH4ST3QoJ9zuP'


@skipif_no_gdrive_readonly
def test_gdrive_file_id_multiple():
    # multiple files with the same name in the same folder
    path = 'MSL/msl-io-testing/f 1/electronics.xlsx'

    with pytest.raises(OSError) as err:
        dr.file_id(path)
    assert 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet' in str(err.value)
    assert GSheets.MIME_TYPE in str(err.value)

    mime_types = {
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': '1aCSP8HU7mAz2hss8dP7IpNz0xJDzWSe1',
        GSheets.MIME_TYPE: '1SdLw6tlh4EaPeDis0pPepzYRBb_mx_i8fOwgODwQKaE',
    }
    for mime, id_ in mime_types.items():
        assert dr.file_id(path, mime_type=mime) == id_


@skipif_no_gdrive_readonly
@skipif_no_gdrive_writeable
def test_gdrive_create_delete_folder():

    # instantiated in read-only mode
    with pytest.raises(HttpError, match=r'[iI]nsufficient'):
        dr.create_folder('TEST')

    u1 = str(uuid.uuid4())
    u2 = str(uuid.uuid4())

    # create (relative to root)
    for folder in [u1, u2 + '/sub-2/a b c']:
        folder_id = dw.create_folder(folder)
        assert dw.folder_id(folder) == folder_id

    # delete
    for folder in [u1, u2 + '/sub-2/a b c', u2 + '/sub-2', u2]:
        dw.delete(dw.folder_id(folder))
        with pytest.raises(OSError, match='Not a valid Google Drive folder'):
            dw.folder_id(folder)

    # create (relative to a parent folder)
    # ID of "MSL/msl-io-testing/f 1/f2/sub folder 3" is "1wLAPHCOphcOITR37b8UB88eFW_FzeNQB"
    u3 = str(uuid.uuid4())
    folder_id = dw.create_folder(u3 + '/a/b/c', parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')
    assert dw.folder_id('MSL/msl-io-testing/f 1/f2/sub folder 3/' + u3 + '/a/b/c') == folder_id

    # these should not raise an error (do not need to assert anything)
    u3_id = dw.folder_id(u3, parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')
    u3_a_id = dw.folder_id('a', parent_id=u3_id)
    u3_a_b_id = dw.folder_id('b', parent_id=u3_a_id)
    u3_a_b_c_id = dw.folder_id('c', parent_id=u3_a_b_id)

    # deleting a folder should also delete the children folders
    dw.delete(u3_id)
    assert dw.is_folder('sub folder 3', parent_id='1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN')
    assert not dw.is_folder(u3 + '/a/b/c', parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')
    assert not dw.is_folder(u3 + '/a/b', parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')
    assert not dw.is_folder(u3 + '/a', parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')
    assert not dw.is_folder(u3, parent_id='1wLAPHCOphcOITR37b8UB88eFW_FzeNQB')

    # create multiple folders with the same name, check exists
    id1 = dw.create_folder('Temporary')
    id2 = dw.create_folder('Temporary')
    assert dw.path(id1) == 'My Drive/Temporary'
    assert dw.path(id2) == 'My Drive/Temporary'
    assert id1 != id2
    assert dw.is_folder('Temporary') is True
    with pytest.raises(OSError, match='^Multiple folders'):
        dw.folder_id('Temporary')
    dw.delete(id1)
    dw.delete(id2)


@skipif_no_gdrive_readonly
def test_gdrive_is_file():
    # relative to the root folder
    assert not dr.is_file('doesnotexist.txt')
    assert not dr.is_file('does/not/exist.txt')
    assert not dr.is_file('MSL')
    assert dr.is_file('MSL/msl-io-testing/unique')
    assert dr.is_file('MSL/msl-io-testing/f 1/electronics.xlsx')
    assert dr.is_file('MSL/msl-io-testing/f 1/electronics.xlsx', mime_type=GSheets.MIME_TYPE)
    if IS_WINDOWS:
        assert not dr.is_file('C:\\Users\\username\\Google Drive\\MSL\\msl-io-testing\\f 1')
        assert dr.is_file(r'MSL\msl-io-testing\f 1\f2\New Text Document.txt')

    # relative to a parent folder
    assert dr.is_file('unique', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C')
    assert dr.is_file('msl-io-testing/unique', folder_id='14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi')
    assert dr.is_file('Single-Photon Generation and Detection.pdf', folder_id='root')
    assert dr.is_file('f 1/f2/New Text Document.txt', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C')
    assert not dr.is_file('f2', folder_id='1mhQ_9iVF5AhbUb7Lq77qzuBbvZr150X9')
    assert not dr.is_file('New Text Document.txt', folder_id='1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C')

    # relative to an invalid parent folder
    with pytest.raises(HttpError):
        dr.is_file('unique', folder_id='INVALID_NCuTWxmABs-w7JenftaLGAG9C')


@skipif_no_sheets_writeable
@skipif_no_gdrive_writeable
def test_gsheets_append():
    sid = sw.create('appending')
    sw.append(None, sid)
    sw.append([], sid)
    sw.append([[]], sid)
    sw.append(0, sid)
    sw.append([1, 2], sid, sheet='Sheet1')
    sw.append([[3, 4, 5, 6], [7, 8, 9]], sid)
    sw.append([[10, 11, 12, 13], [14, 15, 16]], sid, row_major=False)
    sw.append([None, 17], sid)
    assert sw.values(sid) == [
        ['0'],
        ['1', '2'],
        ['3', '4', '5', '6'],
        ['7', '8', '9'],
        ['10', '14'],
        ['11', '15'],
        ['12', '16'],
        ['13'],
        ['', '17']
    ]
    dw.delete(sid)

    sid = sw.create('appending-2', sheet_names=['Appender'])
    sw.append(['a', 'b', 'c'], sid)
    sw.append(['d', 'e', 'f', 'g'], sid, cell='D4')
    sw.append('h', sid, sheet='Appender')
    sw.append([['i', 'j'], ['k', 'l']], sid, cell='B7')
    sw.append([['m', 'n', 'o'], ['p', 'q', 'r']], sid, row_major=False, cell='A13')
    sw.append('s', sid, cell='A1')
    assert sw.values(sid) == [
        ['a', 'b', 'c'],
        ['s'],
        [],
        [],
        ['', '', '', 'd', 'e', 'f', 'g'],
        ['', '', '', 'h'],
        [],
        ['', 'i', 'j'],
        ['', 'k', 'l'],
        [],
        [],
        [],
        [],
        ['m', 'p'],
        ['n', 'q'],
        ['o', 'r']
    ]
    dw.delete(sid)


@skipif_no_sheets_writeable
@skipif_no_gdrive_writeable
def test_gsheets_write():
    sid = sw.create('writing')
    sw.write(None, sid, 'A1')
    sw.write([], sid, 'A1')
    sw.write([[]], sid, 'A1')
    sw.write(0, sid, 'C1')
    sw.write([1, 2], sid, 'A2:B3', sheet='Sheet1')
    sw.write([[3, 4, 5, 6], [7, 8, 9, 10], [11, 12, 13, 14]], sid, 'A4', row_major=False)
    assert sw.values(sid) == [
        ['', '', '0'],
        ['1', '2'],
        [],
        ['3', '7', '11'],
        ['4', '8', '12'],
        ['5', '9', '13'],
        ['6', '10', '14'],
    ]

    values = [list(range(10)), list(range(10, 20)), list(range(20, 30)), list(range(30, 40))]
    sw.write(values, sid, 'A2')
    expected = [['', '', 0]]
    expected.extend(values)
    expected.extend([[5, 9, 13], [6, 10, 14]])
    assert sw.values(sid, value_option=GValueOption.UNFORMATTED, sheet='Sheet1') == expected

    dw.delete(sid)


@skipif_no_sheets_writeable
@skipif_no_gdrive_writeable
def test_gsheets_copy_rename_add_delete():
    id1 = sw.create('spreadsheet1', sheet_names=['a', 'b', 'c'])
    id2 = sw.create('spreadsheet2')

    d_sheet = sw.add_sheets('d', id1)
    assert list(d_sheet.values()) == ['d']
    sheets = sw.add_sheets(['e', 'f', 'g'], id1)
    assert list(sheets.values()) == ['e', 'f', 'g']
    assert sw.sheet_names(id1) == ('a', 'b', 'c', 'd', 'e', 'f', 'g')

    assert sw.sheet_names(id2) == ('Sheet1',)
    bid = sw.copy('b', id1, id2)
    assert sw.sheet_names(id2) == ('Sheet1', 'Copy of b')
    sw.copy(list(d_sheet.keys())[0], id1, id2)
    assert sw.sheet_names(id2) == ('Sheet1', 'Copy of b', 'Copy of d')

    sw.rename_sheet(bid, 'B - B', id2)
    assert sw.sheet_names(id2) == ('Sheet1', 'B - B', 'Copy of d')

    sw.rename_sheet('a', 'different', id1)
    assert sw.sheet_names(id1) == ('different', 'b', 'c', 'd', 'e', 'f', 'g')

    sw.delete_sheets('different', id1)
    assert sw.sheet_names(id1) == ('b', 'c', 'd', 'e', 'f', 'g')
    sw.delete_sheets(bid, id2)
    assert sw.sheet_names(id2) == ('Sheet1', 'Copy of d')
    sw.delete_sheets(['g', list(d_sheet.keys())[0]], id1)
    assert sw.sheet_names(id1) == ('b', 'c', 'e', 'f')

    with pytest.raises(ValueError, match="no sheet named 'invalid'"):
        sw.copy('invalid', id1, id2)

    with pytest.raises(ValueError, match="no sheet named 'invalid'"):
        sw.delete_sheets('invalid', id1)

    # cleanup
    dw.delete(id1)
    dw.delete(id2)


@skipif_no_sheets_writeable
@skipif_no_gdrive_writeable
def test_read_only():
    # Although folders cannot be set to be in read-only mode,
    # test that calling is_read_only() does not raise an error
    fid = dw.create_folder('Read Only Test')
    assert dw.is_read_only(fid) is False
    dw.delete(fid)

    # create a spreadsheet file
    fid = sw.create('Lockable')
    assert dw.is_read_only(fid) is False
    sw.write('test', fid, 'A1', 'Sheet1')
    dw.read_only(fid, True)
    assert dw.is_read_only(fid) is True

    # cannot create a new sheet
    with pytest.raises(HttpError):
        sw.add_sheets('Another Sheet', fid)

    # cannot rename
    with pytest.raises(HttpError):
        dw.rename(fid, 'New Name')

    # cannot delete
    with pytest.raises(RuntimeError, match='Cannot delete'):
        dw.delete(fid)

    # cannot modify a cell
    with pytest.raises(HttpError):
        sw.write('test2', fid, 'A2', 'Sheet1')
    assert sw.values(fid) == [['test']]

    # setting the same value multiple times is okay
    dw.read_only(fid, True)
    dw.read_only(fid, True)
    dw.read_only(fid, True)
    assert dw.is_read_only(fid) is True
    dw.read_only(fid, False)
    dw.read_only(fid, False)
    dw.read_only(fid, False)
    assert dw.is_read_only(fid) is False

    dw.delete(fid)


@skipif_no_gdrive_readonly
def test_gdrive_is_folder():
    # relative to the root folder
    assert not dr.is_folder('doesnotexist')
    assert not dr.is_folder('MSL/msl-io-testing/unique')
    assert dr.is_folder('MSL')
    assert dr.is_folder('MSL/msl-io-testing/f 1/f2/sub folder 3')
    if IS_WINDOWS:
        assert not dr.is_folder('MSL\\msl-io-testing\\f 1\\electronics.xlsx')
        assert dr.is_folder(r'MSL\msl-io-testing\f 1')

    # relative to a parent folder
    assert not dr.is_folder('doesnotexist', parent_id='1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN')
    assert not dr.is_folder('sub folder 3xx', parent_id='1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN')
    assert dr.is_folder('sub folder 3', parent_id='1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN')
    assert dr.is_folder('f2', parent_id='1mhQ_9iVF5AhbUb7Lq77qzuBbvZr150X9')
    assert dr.is_folder('msl-io-testing/f 1/f2', parent_id='14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi')

    # relative to an invalid parent folder
    with pytest.raises(HttpError):
        dr.is_folder('f2', parent_id='INVALID_F5AhbUb7Lq77qzuBbvZr150X9')


@skipif_no_gdrive_readonly
@skipif_no_gdrive_writeable
def test_gdrive_upload():
    temp_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()) + '.py')
    with open(temp_file, mode='wt') as fp:
        fp.write('from msl.io import GDrive')

    # instantiated in read-only mode
    with pytest.raises(HttpError, match=r'[iI]nsufficient'):
        dr.upload(temp_file)

    file_id = dw.upload(
        temp_file,
        folder_id=dw.folder_id('MSL'),
        mime_type='text/x-python'
    )

    path = os.path.join('MSL', os.path.basename(temp_file))
    assert dw.file_id(path, mime_type='text/x-python') == file_id
    assert dw.file_id(path) == file_id
    assert not dw.is_file(path, mime_type='application/x-python-code')

    dw.delete(file_id)
    assert not dw.is_file(path)
    os.remove(temp_file)


@skipif_no_gdrive_readonly
def test_gdrive_download():
    temp_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))

    file_id = dr.file_id('MSL/msl-io-testing/file.txt')

    # cannot be a string IO object
    with pytest.raises(TypeError):
        dr.download(file_id, save_to=io.StringIO())
    if not IS_PYTHON2:  # in Python 2, str and bytes are the same
        with pytest.raises(TypeError):
            dr.download(file_id, save_to=open('junk.txt', mode='wt'))
        os.remove('junk.txt')  # clean up since it got created before the error was raised

    # a BytesIO object
    with io.BytesIO() as buffer:
        dr.download(file_id, save_to=buffer)
        buffer.seek(0)
        assert buffer.read() == b'in "msl-io-testing"'

    # a file handle in 'wb' mode
    with open(temp_file, mode='wb') as fp:
        dr.download(file_id, save_to=fp)
    with open(temp_file, mode='rt') as fp:
        assert fp.read() == 'in "msl-io-testing"'
    os.remove(temp_file)  # clean up

    # do not specify a value for the 'save_to' kwarg
    # therefore the filename is determined from the remote filename
    # and saved to the current working directory
    file_id = dr.file_id('MSL/msl-io-testing/f 1/f2/sub folder 3/file.txt')
    dr.download(file_id)
    with open('file.txt', mode='rt') as fp:
        assert fp.read() == 'in "sub folder 3"'
    os.remove('file.txt')  # clean up

    # save to a specific directory, use the remote filename
    f = os.path.join(tempfile.gettempdir(), 'file.txt')
    if os.path.isfile(f):
        os.remove(f)
    assert not os.path.isfile(f)
    dr.download(file_id, save_to=tempfile.gettempdir())
    with open(f, mode='rt') as fp:
        assert fp.read() == 'in "sub folder 3"'
    os.remove(f)  # clean up

    # save to a specific file
    assert not os.path.isfile(temp_file)
    dr.download(file_id, save_to=temp_file)
    with open(temp_file, mode='rb') as fp:
        assert fp.read() == b'in "sub folder 3"'
    os.remove(temp_file)  # clean up

    # use a callback
    def handler(file):
        assert file.progress() == 1.0
        assert file.total_size == 17
        assert file.resumable_progress == 17
    dr.download(file_id, save_to=temp_file, callback=handler)
    os.remove(temp_file)  # clean up


@skipif_no_gdrive_readonly
@skipif_no_gdrive_writeable
def test_gdrive_empty_trash():
    # instantiated in read-only mode
    with pytest.raises(HttpError, match=r'[iI]nsufficient'):
        dr.empty_trash()
    dw.empty_trash()


@skipif_no_gdrive_readonly
def test_gdrive_path():
    assert dr.path('0AFP6574OTgaaUk9PVA') == 'My Drive'
    assert dr.path('11yaxZH93B0IhQZwfCeo2dXb-Iduh-4dS') == 'My Drive/Single-Photon Generation and Detection.pdf'
    assert dr.path('14GYO5FIKmkjo9aCQGysOsxwkTtpoJODi') == 'My Drive/MSL'
    assert dr.path('1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C') == 'My Drive/MSL/msl-io-testing'
    assert dr.path('1Ua15pRGUH5qoU0c3Ipqrkzi9HBlm3nzqCn5O1IONfCY') == 'My Drive/MSL/msl-io-testing/empty-5'
    assert dr.path('1HG_emhGXBGaR7oS6ftioJOF-xbl1kv41') == 'My Drive/MSL/msl-io-testing/file.txt'
    assert dr.path('1iaLNB_IZNxbFlpy-Z2-22WQGWy4wU395') == 'My Drive/MSL/msl-io-testing/unique'
    assert dr.path('1mhQ_9iVF5AhbUb7Lq77qzuBbvZr150X9') == 'My Drive/MSL/msl-io-testing/f 1'
    assert dr.path('1SdLw6tlh4EaPeDis0pPepzYRBb_mx_i8fOwgODwQKaE') == 'My Drive/MSL/msl-io-testing/f 1/electronics.xlsx'
    assert dr.path('1aCSP8HU7mAz2hss8dP7IpNz0xJDzWSe1') == 'My Drive/MSL/msl-io-testing/f 1/electronics.xlsx'
    assert dr.path('1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN') == 'My Drive/MSL/msl-io-testing/f 1/f2'
    assert dr.path('1qW1QclelxZtJtKMigCgGH4ST3QoJ9zuP') == 'My Drive/MSL/msl-io-testing/f 1/f2/New Text Document.txt'
    assert dr.path('1wLAPHCOphcOITR37b8UB88eFW_FzeNQB') == 'My Drive/MSL/msl-io-testing/f 1/f2/sub folder 3'
    assert dr.path('1CDS3cWDItXB1uLCPGq0uy6OJAngkmNoD') == 'My Drive/MSL/msl-io-testing/f 1/f2/sub folder 3/file.txt'
    assert dr.path('1FwzsFgN7w-HZXOlUAEMVMSOGpNHCj5NXvH6Xl7LyLp4') == 'My Drive/MSL/msl-io-testing/f 1/f2/sub folder 3/lab environment'


@skipif_no_gdrive_writeable
def test_gdrive_copy():
    msl_io_testing_id = '1oB5i-YcNCuTWxmABs-w7JenftaLGAG9C'
    assert dw.path(msl_io_testing_id) == 'My Drive/MSL/msl-io-testing'

    file_txt_id = '1HG_emhGXBGaR7oS6ftioJOF-xbl1kv41'
    assert dw.path(file_txt_id) == 'My Drive/MSL/msl-io-testing/file.txt'

    f2_id = '1NRD4klmRTQDkh5ZfhnhaHc6hDYfklMJN'
    assert dw.path(f2_id) == 'My Drive/MSL/msl-io-testing/f 1/f2'

    # copy to the same folder (do not specify the destination folder)
    cid = dw.copy(file_txt_id)
    assert cid != file_txt_id
    assert dw.path(cid) == 'My Drive/MSL/msl-io-testing/file.txt'
    with pytest.raises(OSError, match='^Multiple files'):
        dw.file_id('file.txt', folder_id=msl_io_testing_id)
    assert dw.is_file('file.txt', folder_id=msl_io_testing_id) is True
    dw.delete(cid)
    assert dw.is_file('file.txt', folder_id=msl_io_testing_id)

    # copy to a different folder
    cid = dw.copy(file_txt_id, f2_id)
    assert cid != file_txt_id
    assert dw.path(cid) == 'My Drive/MSL/msl-io-testing/f 1/f2/file.txt'
    dw.delete(cid)
    assert dw.is_file('file.txt', folder_id=msl_io_testing_id)

    # copy to the same folder (but specify it) and rename
    cid = dw.copy(file_txt_id, msl_io_testing_id, name='new-file.dat')
    assert cid != file_txt_id
    assert dw.path(cid) == 'My Drive/MSL/msl-io-testing/new-file.dat'
    dw.delete(cid)
    assert not dw.is_file('new-file.dat', folder_id=msl_io_testing_id)
    assert dw.is_file('file.txt', folder_id=msl_io_testing_id)

    # copy to a different folder and rename (do not specify an extension)
    cid = dw.copy(file_txt_id, f2_id, name='abc')
    assert cid != file_txt_id
    assert dw.path(cid) == 'My Drive/MSL/msl-io-testing/f 1/f2/abc'
    dw.delete(cid)
    assert not dw.is_file('abc', folder_id=f2_id)
    assert dw.is_file('file.txt', folder_id=msl_io_testing_id)


@skipif_no_gdrive_writeable
def test_gdrive_rename():
    # rename a folder
    fid = dw.create_folder('My Folder')
    assert dw.path(fid) == 'My Drive/My Folder'
    dw.rename(fid, 'Renamed Folder')
    assert dw.path(fid) == 'My Drive/Renamed Folder'

    # rename a file
    file_txt_id = '1HG_emhGXBGaR7oS6ftioJOF-xbl1kv41'
    assert dw.path(file_txt_id) == 'My Drive/MSL/msl-io-testing/file.txt'
    cid = dw.copy(file_txt_id, fid)
    assert dw.path(cid) == 'My Drive/Renamed Folder/file.txt'
    dw.rename(cid, 'renamed file.txt')
    assert dw.path(cid) == 'My Drive/Renamed Folder/renamed file.txt'

    # cleanup
    dw.delete(fid)


@skipif_no_gdrive_writeable
def test_gdrive_move():
    # move a folder
    fid = dw.create_folder('X/Y/Z')
    assert dw.path(fid) == 'My Drive/X/Y/Z'
    dw.move(fid, 'root')
    assert dw.path(fid) == 'My Drive/Z'
    dw.move(fid, dw.folder_id('My Drive/X'))
    assert dw.path(fid) == 'My Drive/X/Z'

    # move a file (first create a copy)
    cid = dw.copy(dw.file_id('MSL/msl-io-testing/file.txt'), folder_id='root', name='copied.txt')
    assert dw.path(cid) == 'My Drive/copied.txt'
    dw.move(cid, fid)
    assert dw.path(cid) == 'My Drive/X/Z/copied.txt'

    # cleanup
    dw.delete(dw.folder_id('X'))


@skipif_no_gmail
def test_gmail_profile():
    profile = gmail.profile()
    assert profile['email_address'].endswith('@gmail.com')
    assert profile['messages_total'] > 1
    assert profile['threads_total'] > 1
    if IS_PYTHON2:
        assert isinstance(profile['history_id'], unicode)
    else:
        assert isinstance(profile['history_id'], str)
